# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.common."""

import copy
import logging

from vizier._src.pyvizier.shared import common
from google.protobuf import any_pb2
from google.protobuf import duration_pb2
from absl.testing import absltest


class MetadataGetClsTest(absltest.TestCase):

  def test_get_proto(self):
    duration = duration_pb2.Duration(seconds=60)
    anyproto = any_pb2.Any()
    anyproto.Pack(duration)
    metadata = common.Metadata(duration=duration, any=anyproto)

    self.assertEqual(
        metadata.get_proto('duration', cls=duration_pb2.Duration), duration)
    self.assertEqual(
        metadata.get_proto('any', cls=duration_pb2.Duration), duration)
    self.assertEqual(
        metadata.get('duration', cls=duration_pb2.Duration), duration)
    self.assertEqual(metadata.get('any', cls=duration_pb2.Duration), duration)

  def test_get_int(self):
    metadata = common.Metadata({'string': '30', 'int': '60'})
    self.assertEqual(metadata.get('string', cls=int), 30)
    self.assertEqual(metadata.get('int', cls=int), 60)
    self.assertEqual(metadata.get('badkey', 1, cls=int), 1)


class MetadataNamespaceTest(absltest.TestCase):

  def test_basic(self):
    ns0 = common.Namespace()
    self.assertEmpty(ns0)
    self.assertEqual(str(ns0), '')
    self.assertEqual(ns0.encode(), '')
    self.assertEqual(ns0, common.Namespace.decode(''))
    n1t = common.Namespace(('aerer',))
    self.assertLen(n1t, 1)
    n1 = common.Namespace.decode('a78')
    self.assertLen(n1, 1)
    self.assertEqual(str(n1), ':a78')
    n2 = common.Namespace(('a78', 'bfe'))
    self.assertLen(n2, 2)
    n2s1 = common.Namespace.decode('a78:bfe')
    self.assertLen(n2s1, 2)
    self.assertEqual(n2.encode(), n2s1.encode())
    n2s2 = common.Namespace.decode(':a78:bfe')
    self.assertLen(n2s2, 2)
    self.assertEqual(n2.encode(), n2s2.encode())
    self.assertEqual(n2, n2s2)
    self.assertEqual(n2s1, n2s2)
    ns = common.Namespace(('a', 'b'))
    self.assertLen(ns, 2)
    self.assertEqual(tuple(ns), ('a', 'b'))
    self.assertEqual(str(ns), ':a:b')
    self.assertEqual(ns.encode(), ':a:b')

  def test_escape(self):
    s1 = 'a\\:A'
    ns1 = common.Namespace.decode(s1)
    self.assertLen(ns1, 1)
    self.assertEqual(str(ns1), ':a\\:A')
    self.assertEqual(ns1.encode(), ':' + s1)
    self.assertEqual(common.Namespace.decode(ns1.encode()), ns1)
    #
    s2 = 'b:B'
    ns2 = common.Namespace.decode(s2)
    self.assertLen(ns2, 2)
    self.assertEqual(str(ns2), ':' + s2)
    self.assertEqual(ns2.encode(), ':' + s2)
    self.assertEqual(common.Namespace.decode(ns2.encode()), ns2)
    #
    s1e1 = ':b\\B'
    ns1e1 = common.Namespace.decode(s1e1)
    self.assertLen(ns1e1, 1)
    self.assertEqual(ns1e1.encode(), s1e1)
    self.assertEqual(common.Namespace.decode(ns1e1.encode()), ns1e1)
    ns1e2 = common.Namespace((s1e1.lstrip(':'),))
    self.assertLen(ns1e2, 1)
    self.assertEqual(ns1e2.encode(), s1e1)
    self.assertEqual(ns1e2, ns1e1)
    self.assertEqual(common.Namespace.decode(ns1e2.encode()), ns1e2)
    #
    s1c = r':b\:B'
    ns1c = common.Namespace.decode(s1c)
    self.assertLen(ns1c, 1)
    # Initial colon is harmlessly removed.
    self.assertEqual(ns1c.encode(), s1c)
    self.assertEqual(common.Namespace.decode(ns1c.encode()), ns1c)
    self.assertEqual(common.Namespace(('b:B',)), ns1c)


class MetadataTest(absltest.TestCase):

  def create_test_metadata(self):
    md = common.Metadata({'bar': 'bar_v'}, foo='foo_v')
    md.ns('Name').update(foo='Name_foo_v', baz='Name_baz_v')
    return md

  def test_empty_namespaces(self):
    md = common.Metadata()
    self.assertEmpty(list(md.namespaces()))
    md = common.Metadata().ns('ns')
    self.assertEmpty(list(md.namespaces()))

  def test_nonempty_namespaces(self):
    mm = self.create_test_metadata()
    self.assertLen(mm.namespaces(), 2)

  def test_getters_are_consistent_when_item_is_in_dict(self):
    mm = self.create_test_metadata()
    self.assertEqual(mm['foo'], 'foo_v')
    self.assertEqual(mm.get('foo'), 'foo_v')

  def test_getters_are_consistent_when_item_is_not_in_dict(self):
    mm = self.create_test_metadata()
    self.assertIsNone(mm.get('baz'))
    with self.assertRaises(KeyError):
      _ = mm['baz']

  def test_separator_is_not_allowed_as_keys_after_init(self):
    mm = self.create_test_metadata()
    with self.assertRaises(KeyError):
      _ = mm['Name_foo']

  def test_namespace_works_as_intended(self):
    mm = self.create_test_metadata()
    self.assertEqual(mm.ns('Name')['foo'], 'Name_foo_v')
    self.assertIsNone(mm.ns('Name').get('bar'))

    mm_name = mm.ns('Name')
    self.assertEqual(mm_name['foo'], 'Name_foo_v')
    self.assertIsNone(mm_name.get('bar'))
    self.assertEqual(mm.ns('Name')['foo'], 'Name_foo_v')

  def test_create_new_namespace(self):
    # Calling ns() with an unexisting namespace should work fine.
    mm = self.create_test_metadata()
    mm.ns('NewName')['foo'] = 'NewName_foo_v'
    self.assertEqual(mm.ns('NewName')['foo'], 'NewName_foo_v')
    self.assertIsNone(mm.ns('NewName').get('bar'))

  def test_changing_namespace_copies_reference(self):
    mm = self.create_test_metadata()
    # Calling ns() copies by reference so any changes to the returned Metadata
    # object is reflected in the original object.
    mm_in_namespace = mm.ns('Name')
    mm_in_namespace['foofoo'] = 'Name_foofoo_v'
    self.assertEqual(mm.ns('Name')['foofoo'], 'Name_foofoo_v')

  def test_iterator(self):
    mm = self.create_test_metadata()
    self.assertSequenceEqual(list(mm.keys()), ['bar', 'foo'])
    self.assertSequenceEqual(
        list(mm.ns('Name').values()), ['Name_foo_v', 'Name_baz_v'])
    self.assertLen(list(mm.items()), 2)

  def test_all_items(self):
    mm = self.create_test_metadata()
    sorted_tuples = list(sorted(mm.all_items()))
    logging.info('sorted_tuples: %s', sorted_tuples)
    self.assertLen(sorted_tuples, 4)
    self.assertSequenceEqual(
        sorted_tuples, [(common.Namespace([]), 'bar', 'bar_v'),
                        (common.Namespace([]), 'foo', 'foo_v'),
                        (common.Namespace(['Name']), 'baz', 'Name_baz_v'),
                        (common.Namespace(['Name']), 'foo', 'Name_foo_v')])

  def test_repr_str(self):
    mm = self.create_test_metadata()
    self.assertNotEmpty(str(mm), '')
    self.assertNotEmpty(repr(mm), repr(''))

  def test_update(self):
    md = common.Metadata(foo='foo_v')
    md.ns('Name').update(foo='Name_foo_v', baz='Name_baz_v')

    md2 = common.Metadata()
    md2.ns('Name').update(foo='Name_foo_v2', bar='Name_bar_v2')

    md.ns('Name').update(md2.ns('Name'))

    self.assertLen(md.ns('Name'), 3)
    self.assertIn('bar', md.ns('Name'))

  def test_copy(self):
    # There's no useful distinction to be made between copy.copy() and
    # copy.deepcopy().
    mm = common.Metadata().ns('ns1')
    mm.update(foo='bar')
    mm_copy = copy.copy(mm)
    mm_deepcopy = copy.deepcopy(mm)
    # Check that copies match.
    self.assertEqual(mm['foo'], 'bar')
    self.assertEqual(mm_copy['foo'], 'bar')
    self.assertEqual(mm_deepcopy['foo'], 'bar')
    self.assertEqual(mm_deepcopy.namespaces(), mm.namespaces())
    self.assertEqual(mm_copy.namespaces(), mm.namespaces())
    # Check that the  deep copy is disconnected.
    mm_deepcopy['nerf'] = 'gleep'
    with self.assertRaises(KeyError):
      mm['nerf']  # pylint: disable=pointless-statement
    with self.assertRaises(KeyError):
      mm_copy['nerf']  # pylint: disable=pointless-statement
    # Check that the shallow copy shares the metadata store with the original.
    mm_copy['blip'] = 'tonk'
    self.assertEqual(mm['blip'], mm_copy['blip'])
    # ... but no sharing with the deep copy.
    with self.assertRaises(KeyError):
      mm_deepcopy['blip']  # pylint: disable=pointless-statement
    # Here's a test for a specific bug, where Metadata._store is improperly
    # disconnected from Metadata._stores.
    mx = common.Metadata()
    copy.copy(mx).ns('A')['a'] = 'Aa'
    self.assertEqual(mx.ns('A')['a'], 'Aa')

  def test_construction(self):
    # Test with iterables.
    m0i = common.Namespace([])
    self.assertEmpty(m0i)
    m0d = common.Namespace.decode('')
    self.assertEmpty(m0d)
    self.assertEqual(m0d, m0i)
    m1i = common.Namespace(['abc'])
    self.assertLen(m1i, 1)
    self.assertEqual(m1i, common.Namespace(tuple(m1i)))
    self.assertEqual(m1i, common.Namespace.decode(m1i.encode()))
    m2i = common.Namespace(['abc', 'def'])
    self.assertLen(m2i, 2)
    self.assertEqual(m2i, common.Namespace(tuple(m2i)))
    self.assertEqual(m2i, common.Namespace.decode(m2i.encode()))
    m3i = common.Namespace(['abc', 'de:f'])
    self.assertLen(m3i, 2)
    self.assertEqual(m3i, common.Namespace(tuple(m3i)))
    self.assertEqual(m3i, common.Namespace.decode(m3i.encode()))
    # Test with strings.
    m1sc = common.Namespace.decode(':abc')
    self.assertLen(m1sc, 1)
    self.assertEqual(m1sc, common.Namespace(tuple(m1sc)))
    self.assertEqual(m1sc, common.Namespace.decode(m1sc.encode()))
    m1s = common.Namespace.decode('abc')
    self.assertLen(m1s, 1)
    self.assertEqual(m1s, common.Namespace(tuple(m1s)))
    self.assertEqual(m1s, common.Namespace.decode(m1s.encode()))
    m2s = common.Namespace.decode('abc:def')
    self.assertLen(m2s, 2)
    self.assertEqual(m2s, common.Namespace(tuple(m2s)))
    self.assertEqual(m2s, common.Namespace.decode(m2s.encode()))
    m3s = common.Namespace.decode('abc:de\\f')
    self.assertLen(m3s, 2)
    self.assertEqual(m3s, common.Namespace(tuple(m3s)))
    self.assertEqual(m3s, common.Namespace.decode(m3s.encode()))

  def test_assign_proto(self):
    m0 = common.Metadata()
    m0['foo'] = duration_pb2.Duration(seconds=60)
    dur_out = m0.get('foo', cls=duration_pb2.Duration)
    logging.info('dur_out= %s', dur_out)
    self.assertIsNotNone(dur_out)
    self.assertIsInstance(dur_out, duration_pb2.Duration)
    if dur_out is not None:
      self.assertEqual(dur_out.seconds, 60)

  def test_get_or_error(self):
    m0 = common.Metadata({'foo': 33, 'bar': 'Z'})
    m0['gleep'] = duration_pb2.Duration(seconds=60)
    self.assertEqual(m0.get_or_error('foo', cls=int), 33)
    self.assertEqual(m0.get_or_error('bar'), 'Z')
    dur_out = m0.get_or_error('gleep', cls=duration_pb2.Duration)
    self.assertIsInstance(dur_out, duration_pb2.Duration)
    self.assertEqual(dur_out.seconds, 60)

  def test_startswith(self):
    m1 = common.Namespace(['aa', 'bb'])
    self.assertTrue(m1.startswith(common.Namespace(['aa'])))
    self.assertTrue(m1.startswith(common.Namespace(['aa', 'bb'])))
    self.assertTrue(m1.startswith(m1))
    self.assertTrue(m1.startswith(common.Namespace(tuple(m1))))
    self.assertFalse(m1.startswith(common.Namespace(['bb'])))
    self.assertFalse(m1.startswith(common.Namespace(['aa', 'bb', 'cc'])))
    self.assertFalse(m1.startswith(common.Namespace(['bb', 'bb'])))
    self.assertFalse(m1.startswith(common.Namespace(['aa', 'aa'])))

  def test_subnamespace(self):
    mm = common.Metadata()
    mm.ns('ns1')['foo'] = 'bar'
    mm.ns('ns2')['foo'] = 'bar'
    mm.ns('ns1').ns('ns11')['foo'] = 'bar'
    mm.ns('ns1').ns('ns:11')['gleep'] = 'nerf'

    self.assertSequenceEqual(mm.subnamespaces(), [
        common.Namespace(['ns1']),
        common.Namespace(['ns2']),
        common.Namespace(['ns1', 'ns11']),
        common.Namespace(['ns1', 'ns:11']),
    ])
    self.assertSequenceEqual(
        mm.ns('ns1').subnamespaces(), [
            common.Namespace([]),
            common.Namespace(['ns11']),
            common.Namespace(['ns:11'])
        ])
    self.assertSequenceEqual(mm.ns('ns2').subnamespaces(), [common.Namespace()])
    self.assertSequenceEqual(mm.ns('ns3').subnamespaces(), [])

  def test_namespace_add(self):
    n0 = common.Namespace()
    self.assertEmpty(n0)
    self.assertEqual(n0 + (), common.Namespace([]))
    self.assertEqual(n0 + ('ab',), common.Namespace([
        'ab',
    ]))
    self.assertEqual(n0 + ('a:b',), common.Namespace(['a:b']))
    self.assertEqual(n0 + ('a:b',), common.Namespace(['a:b']))
    self.assertEqual(n0 + ('ab', 'cd'), common.Namespace(['ab', 'cd']))
    n1 = common.Namespace(['xy'])
    self.assertLen(n1, 1)
    self.assertEqual(n1 + ('ab',), common.Namespace(['xy', 'ab']))
    self.assertEqual(n1 + ('a:b',), common.Namespace(['xy', 'a:b']))
    self.assertEqual(n1 + ('a:b',), common.Namespace(['xy', 'a:b']))
    n2 = common.Namespace(['xy', 'zw'])
    self.assertLen(n2, 2)
    self.assertLen(n2 + ('ab',), 3)
    self.assertEqual(n2 + ('ab',), common.Namespace(['xy', 'zw', 'ab']))
    self.assertLen(n2 + ('ab', 'cd'), 4)
    self.assertEqual(n2 + ('ab', 'cd'), common.Namespace.decode('xy:zw:ab:cd'))

  def test_metadata_attach(self):
    # Set up a metadata tree.
    mm = common.Metadata()
    mm.ns('ns1').ns('ns:11').update(foo='bar')
    mm.ns('ns1').ns('ns12').update(foo='gleep')
    mm.ns('ns1').update(foo='nerf')
    mm.ns('ns|').update(foo='pag')
    # Attach that metadata tree to a branch of an empty tree.
    m1 = common.Metadata()
    m1.ns('ns0').ns('ns00').attach(mm)
    self.assertEmpty(m1.abs_ns())
    self.assertEqual(m1.ns('ns0').ns('ns00'), mm)
    self.assertEqual(m1.abs_ns(['ns0', 'ns00', 'ns1', 'ns:11'])['foo'], 'bar')
    self.assertEqual(m1.abs_ns(['ns0', 'ns00', 'ns1', 'ns12'])['foo'], 'gleep')
    self.assertEqual(m1.abs_ns(['ns0', 'ns00', 'ns1'])['foo'], 'nerf')
    self.assertEqual(m1.abs_ns(['ns0', 'ns00', 'ns|'])['foo'], 'pag')
    # Attach just part of $mm to a branch of a new, empty tree.
    m2 = common.Metadata()
    m2.ns('nsX').attach(mm.ns('ns1'))
    self.assertEqual(m2.abs_ns(['nsX', 'ns:11'])['foo'], 'bar')
    self.assertEqual(m2.abs_ns(['nsX', 'ns12'])['foo'], 'gleep')
    self.assertEqual(m2.abs_ns(['nsX'])['foo'], 'nerf')
    # Check that attach() overwrites key collisions, but preserves other data.
    m3 = common.Metadata()
    m3['foo'] = 'Y'  # This will be overwritten.
    m3['z'] = 'Z'  # This will not be overwritten.
    m3.attach(mm.ns('ns1').ns('ns:11'))
    self.assertEqual(m3['z'], 'Z')
    self.assertEqual(m3['foo'], 'bar')

  def test_metadata_items_by_cls(self):
    mm = common.Metadata(foo='bar', nerf='gleep')
    mm['aaa'] = duration_pb2.Duration(seconds=5)
    mm.ns('ns1')['foo'] = 'bar1'
    test1 = list(sorted(mm.items_by_cls(cls=str)))
    self.assertLen(test1, 2)
    self.assertEqual(test1[0], ('foo', 'bar'))
    self.assertEqual(test1[1], ('nerf', 'gleep'))
    test2 = list(sorted(mm.items_by_cls(cls=duration_pb2.Duration)))
    self.assertLen(test2, 1)
    self.assertEqual(test2[0][1].seconds, 5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""L-BFGS-B Strategy optimizer."""

from typing import Callable, Optional, Union

import attr
from flax import struct
import jax
import jax.numpy as jnp
import numpy as np
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier.jax import optimizers
from vizier.pyvizier import converters


def _optimizer_to_model_input(
    x: jax.Array,
    n_features: jax.Array,
) -> types.ModelInput:
  x_cat = jnp.zeros(x.shape[:-1] + (0,), dtype=types.INT_DTYPE)
  return types.ModelInput(
      continuous=vb.optimizer_to_model_input_single_array(x, n_features),
      categorical=vb.optimizer_to_model_input_single_array(
          x_cat, jnp.zeros([])
      ),
  )


@struct.dataclass
class LBFGSBOptimizer:
  """L-BFGS-B optimizer."""

  n_feature_dimensions: types.ContinuousAndCategorical[jax.Array]
  n_feature_dimensions_with_padding: types.ContinuousAndCategorical[int] = (
      struct.field(pytree_node=False)
  )
  # Number of parallel runs of L-BFGS-B.
  random_restarts: int = struct.field(pytree_node=False, default=25)
  # Number of iterations for each L-BFGS-B run.
  maxiter: int = struct.field(pytree_node=False, default=50)

  def __post_init__(self):
    if self.n_feature_dimensions_with_padding.categorical > 0:
      raise ValueError("LBFGSBOptimizer doesn't support categorical features.")

  # TODO: Remove score_fn argument.
  # pylint: disable=g-bare-generic
  def __call__(
      self,
      score_fn: Union[vb.ArrayScoreFunction, vb.ParallelArrayScoreFunction],
      *,
      score_with_aux_fn: Optional[Callable] = None,
      count: Optional[int] = 1,
      prior_features: Optional[types.ModelInput] = None,
      n_parallel: Optional[int] = None,
      seed: Optional[int] = None,
  ) -> vb.VectorizedStrategyResults:
    """Optimize a continuous objective function using L-BFGS-B.

    Arguments:
      score_fn: A callback that expects 2D Array with dimensions (batch_size,
        features_count) or a 3D array with dimensions (batch_size, n_parallel,
        features_count), and a random seed, and returns a 1D Array
        (batch_size,).
      score_with_aux_fn: A callback similar to score_fn but additionally returns
        an array tree.
      count: The number of suggestions to generate.
      prior_features: Completed trials to be used for knowledge transfer.
        Currently ignored.
      n_parallel: This arg should be specified if a parallel acquisition
        function (e.g. qEI, qUCB) is used, which evaluates a set of points of
        this size to a single value. Pass `None` when optimizing acquisition
        functions that evaluate a single point. (Note that `num_parallel=1` and
        `num_parallel=None` will each return a single suggestion, though they
        are different in that `num_parallel=1` indicates that the acquisition
        function consumes the two rightmost dimensions of the input while
        `num_parallel=None` indicates that only the rightmost dimension is
        consumed.)
      seed: The seed to use in the random generator.

    Returns:
      The best trials found in the optimization.
    """
    del prior_features
    seed = jax.random.PRNGKey(0) if seed is None else seed
    if n_parallel and (count or 0) > 1:
      # Note that we can't distinguish between 'BatchArrayScoreFunction' and
      # 'ParallelArrayScoreFunction' using 'isinstance' as they both have the
      # same function names.
      raise ValueError(
          "LBFGSBOptimizer doesn't support batch of batches (count > 1 is"
          " disallowed when num_parallel_candidates is set)."
      )

    parallel_dim = n_parallel or 1
    optimize = optimizers.JaxoptScipyLbfgsB(
        optimizers.LbfgsBOptions(maxiter=self.maxiter)
    )

    score_rng, init_seed, optim_seed = jax.random.split(seed, num=3)

    score_with_aux = score_with_aux_fn
    if score_with_aux is None:
      score_with_aux = lambda *args: (score_fn(*args), dict())

    # We'll assume score_fn is differentiable and use the same acquisition
    # function seed at every call.
    # We also wrap the score function because the convention for optimizers is
    # to maximize, while L-BFGS-B always minimizes.
    def _opt_score_fn(x):
      score, aux = score_with_aux(
          _optimizer_to_model_input(
              x,
              self.n_feature_dimensions.continuous,
          ),
          score_rng,
      )
      if n_parallel is None:
        score = jnp.squeeze(score, axis=-1)
      return -score, aux

    feature_shape = (
        parallel_dim,
        self.n_feature_dimensions_with_padding.continuous,
    )

    def setup(rng):
      return jax.random.uniform(rng, shape=feature_shape)

    # Constraints are [0, 1].
    constraints = sp.Constraint(
        bounds=(np.zeros(feature_shape), np.ones(feature_shape))
    )

    new_features, _ = optimize(
        jax.vmap(setup)(jax.random.split(init_seed, self.random_restarts)),
        _opt_score_fn,
        optim_seed,
        constraints=constraints,
        best_n=count,
    )
    loss_val, aux = _opt_score_fn(new_features)
    new_rewards = -loss_val

    dimension_is_missing = (
        jnp.arange(self.n_feature_dimensions_with_padding.continuous)
        >= self.n_feature_dimensions.continuous
    )
    new_features_model_input = vb.VectorizedOptimizerInput(
        continuous=jnp.where(
            dimension_is_missing, jnp.zeros_like(new_features), new_features
        ),
        categorical=jnp.zeros(
            new_features.shape[:-1] + (0,), dtype=types.INT_DTYPE
        ),
    )
    return vb.VectorizedStrategyResults(
        new_features_model_input,
        new_rewards,
        aux,
    )


@attr.define
class LBFGSBOptimizerFactory:
  """LBFGSB strategy optimizer factory."""

  random_restarts: int = 25
  maxiter: int = 100

  def __call__(
      self,
      converter: converters.TrialToModelInputConverter,
  ) -> LBFGSBOptimizer:
    """Generates a new LBFGSBOptimizer object."""
    n_feature_dimensions = types.ContinuousAndCategorical(
        jnp.array(len(converter.output_specs.continuous)),
        jnp.array(len(converter.output_specs.categorical)),
    )
    empty_features = converter.to_features([])
    n_feature_dimensions_with_padding = types.ContinuousAndCategorical[int](
        empty_features.continuous.shape[-1],
        empty_features.categorical.shape[-1],
    )
    return LBFGSBOptimizer(
        n_feature_dimensions=n_feature_dimensions,
        n_feature_dimensions_with_padding=n_feature_dimensions_with_padding,
        random_restarts=self.random_restarts,
        maxiter=self.maxiter,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utilities for GP Bandit."""

import jax
from jax import numpy as jnp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types


def stochastic_process_model_loss_fn(
    params: types.ParameterDict,
    model: sp.StochasticProcessModel,
    data: types.ModelData,
    normalize: bool = False,
):
  """Loss function for a stochastic process model."""
  gp, mutables = model.apply(
      {'params': params},
      data.features,
      mutable=['losses', 'predictive'],
  )
  labels = data.labels.padded_array
  if len(gp.event_shape) == 1 and labels.shape[-1] == 1:
    labels = jnp.squeeze(data.labels.padded_array, axis=-1)
  loss = -gp.log_prob(
      labels,
      is_missing=data.labels.is_missing[0],
  ) + jax.tree_util.tree_reduce(jnp.add, mutables['losses'])
  if normalize:
    loss /= data.labels.original_shape[0]
  return loss, dict()


def stochastic_process_model_setup(
    key: jax.Array,
    model: sp.StochasticProcessModel,
    data: types.ModelData,
):
  """Setup function for a stochastic process model."""
  return model.init(key, data.features)['params']



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Converters that use JAX."""

from typing import Sequence, Tuple

import attr
import attrs
import jax
import numpy as np
from vizier import pyvizier as vz
from vizier._src.jax import types as vt
from vizier.pyvizier.converters import core
from vizier.pyvizier.converters import padding


# TODO: Slated for deprecation.
class PaddedTrialToArrayConverter:
  """Converts trials to arrays and pads / masks them."""

  _experimental_override = 'I am aware that this code may break at any point.'

  def __init__(
      self,
      impl: core.TrialToArrayConverter,
      experimental_override: str = '',
      *,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(
          num_trials=padding.PaddingType.NONE,
          num_features=padding.PaddingType.NONE,
      ),
  ):
    """SHOULD NOT BE USED! Use factory classmethods e.g. from_study_config."""

    if experimental_override != self._experimental_override:
      raise ValueError(
          'Set "experimental_override" if you want to call __init__ directly. '
          'Otherwise, use TrialToArrayConverter.from_study_config.'
      )
    self._impl = impl
    self._padding_schedule = padding_schedule

  def to_features(self, trials: Sequence[vz.Trial]) -> vt.PaddedArray:
    """Returns the features array with dimension: (n_trials, n_features)."""
    # Pad up the features.
    features = self._impl.to_features(trials)
    return self._padding_schedule.pad_features(features)

  def to_labels(self, trials: Sequence[vz.Trial]) -> vt.PaddedArray:
    """Returns the labels array with dimenion: (n_trials, n_metrics)."""
    # Pad up the labels.
    labels = self._impl.to_labels(trials)
    return self._padding_schedule.pad_labels(labels)

  def to_xy(
      self, trials: Sequence[vz.Trial]
  ) -> Tuple[vt.PaddedArray, vt.PaddedArray]:
    return self.to_features(trials), self.to_labels(trials)

  def to_parameters(self, arr: np.ndarray) -> Sequence[vz.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    # Undo the padding of the parameters in this array.
    if self._padding_schedule is not None:
      # Get the original shape for the features without padding and
      # remove them.
      n_features = sum([spec.num_dimensions for spec in self.output_specs])
      arr = arr[..., :n_features]
    return self._impl.to_parameters(arr)

  @classmethod
  def from_study_config(
      cls,
      study_config: vz.ProblemStatement,
      *,
      scale: bool = True,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(
          num_trials=padding.PaddingType.NONE,
          num_features=padding.PaddingType.NONE,
      ),
      pad_oovs: bool = True,
      max_discrete_indices: int = 0,
      flip_sign_for_minimization_metrics: bool = True,
      dtype=np.float64,
  ) -> 'PaddedTrialToArrayConverter':
    """From study config.

    Args:
      study_config:
      scale: If True, scales the parameters to [0, 1] range.
      padding_schedule: Pads features and labels according to a padding
        schedule. This is to reduce the number of shapes a designer may see, and
        thus reduce JIT retracing.
      pad_oovs: If True, add an extra dimension for out-of-vocabulary values for
        non-CONTINUOUS parameters.
      max_discrete_indices: For DISCRETE and INTEGER parameters that have more
        than this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value. Note this
        default is different from the default in DefaultModelInputConverter.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      dtype: dtype

    Returns:
      PaddedTrialToArrayConverter
    """
    converter = core.TrialToArrayConverter.from_study_config(
        study_config,
        scale=scale,
        pad_oovs=pad_oovs,
        max_discrete_indices=max_discrete_indices,
        flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
        dtype=dtype,
    )
    return cls(
        converter, cls._experimental_override, padding_schedule=padding_schedule
    )

  @property
  def output_specs(self) -> Sequence[core.NumpyArraySpec]:
    return self._impl.output_specs

  @property
  def metric_specs(self) -> Sequence[vz.MetricInformation]:
    return self._impl.metric_specs

  @property
  def padding_schedule(self) -> padding.PaddingSchedule:
    return self._padding_schedule


@attrs.define
class TrialToModelInputConverter:
  """Converts trials to arrays and pads / masks them."""

  _impl: 'TrialToContinuousAndCategoricalConverter'
  _padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule()

  @classmethod
  def from_problem(
      cls,
      problem: vz.ProblemStatement,
      *,
      scale: bool = True,
      max_discrete_indices: int = 0,
      flip_sign_for_minimization_metrics: bool = True,
      dtype=np.float64,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(),
  ) -> 'TrialToModelInputConverter':
    """Create a new instance from ProblemStatement.

    Args:
      problem:
      scale: If True, scales the parameters to [0, 1] range.
      max_discrete_indices: For DISCRETE and INTEGER parameters that have more
        than this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value. Note this
        default is different from the default in DefaultModelInputConverter.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      dtype: dtype
      padding_schedule:

    Returns:
      TrialToContinuousAndCategoricalConverter
    """

    def create_input_converter(parameter):
      return core.DefaultModelInputConverter(
          parameter,
          scale=scale,
          max_discrete_indices=max_discrete_indices,
          onehot_embed=False,
          float_dtype=dtype,
      )

    def create_output_converter(metric):
      return core.DefaultModelOutputConverter(
          metric,
          flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
          dtype=dtype,
      )

    sc = problem  # alias, to keep pylint quiet in the next line.
    converter = core.DefaultTrialConverter(
        [create_input_converter(p) for p in sc.search_space.parameters],
        [create_output_converter(m) for m in sc.metric_information],
    )
    return cls(
        TrialToContinuousAndCategoricalConverter(converter),
        padding_schedule=padding_schedule,
    )

  def to_features(self, trials: Sequence[vz.Trial]) -> vt.ModelInput:
    """Returns the features array with dimension: (n_trials, n_features)."""
    # Pad up the features.
    features = self._impl.to_features(trials)
    return jax.tree_util.tree_map(self._padding_schedule.pad_features, features)

  def to_labels(self, trials: Sequence[vz.Trial]) -> vt.PaddedArray:
    """Returns the labels array with dimenion: (n_trials, n_metrics)."""
    # Pad up the labels.
    labels = self._impl.to_labels(trials)
    return self._padding_schedule.pad_labels(labels)

  def to_xy(self, trials: Sequence[vz.Trial]) -> vt.ModelData:
    return vt.ModelData(self.to_features(trials), self.to_labels(trials))

  def to_parameters(self, arr: vt.ModelInput) -> Sequence[vz.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    # Undo the padding of the parameters in this array.
    unpadded = vt.ContinuousAndCategoricalArray(
        arr.continuous.unpad(), arr.categorical.unpad()
    )
    return self._impl.to_parameters(unpadded)

  def to_trials(self, data: vt.ModelData) -> Sequence[vz.Trial]:
    unpadded_features = vt.ContinuousAndCategoricalArray(
        data.features.continuous.unpad(), data.features.categorical.unpad()
    )
    unpadded_labels = data.labels.unpad()

    return self._impl.to_trials(unpadded_features, unpadded_labels)

  @property
  def output_specs(self):  # TODO: Add back pytype
    return self._impl.output_specs

  @property
  def metric_specs(self) -> Sequence[vz.MetricInformation]:
    return self._impl.metric_specs


@attr.define
class TrialToContinuousAndCategoricalConverter:
  """EXPERIMENTAL: A converter that returns `ContinuousAndCategoricalArray`s.

  Unlike TrialtoNumpyDict converters, `to_features` returns a structure of
  floating and integer data. CATEGORICAL and DISCRETE parameters are encoded as
  integers and not one-hot embedded.

  IMPORTANT: Use a factory method (currently, there is one: `from_study_config`)
  instead of `__init__`.

  WARNING: This class is not exchangable with `TrialToNumpyDict`, and does not
  have functions that return shape or metric informations. Use it at your own
  risk.
  """

  _impl: core.DefaultTrialConverter

  def to_features(
      self, trials: Sequence[vz.Trial]
  ) -> vt.ContinuousAndCategoricalArray:
    """Returns a structure of arrays with first dimension `n_trials`."""
    features = self._impl.to_features(trials)

    continuous = []
    categorical = []
    for converter, v in zip(self._impl.parameter_converters, features.values()):
      spec = converter.output_spec
      if spec.type == core.NumpyArraySpecType.CONTINUOUS:
        continuous.append(v)
      elif spec.type == core.NumpyArraySpecType.DISCRETE:
        categorical.append(v)
      else:
        raise NotImplementedError(
            f'Expected spec to be CONTINUOUS or DISCRETE, saw {spec.type}'
        )
    if continuous:
      continuous_array = np.concatenate(continuous, axis=-1)
    else:
      continuous_array = np.zeros([len(trials), 0], dtype=np.float64)
    if categorical:
      categorical_array = np.concatenate(categorical, axis=-1).astype(
          vt.INT_DTYPE
      )
    else:
      categorical_array = np.zeros([len(trials), 0], dtype=vt.INT_DTYPE)
    return vt.ContinuousAndCategoricalArray(continuous_array, categorical_array)

  @property
  def dtype(self) -> vt.ContinuousAndCategorical[np.dtype]:
    empty_features = self.to_features([])
    return vt.ContinuousAndCategorical(
        empty_features.continuous.dtype, empty_features.categorical.dtype
    )

  def to_labels(self, trials: Sequence[vz.Trial]) -> np.ndarray:
    """Returns the labels array with dimenion: (n_trials, n_metrics)."""
    return core.dict_to_array(self._impl.to_labels(trials))

  def to_xy(
      self, trials
  ) -> Tuple[vt.ContinuousAndCategoricalArray, np.ndarray]:
    return self.to_features(trials), self.to_labels(trials)

  def _to_dict_2d_arrays(
      self, feat: vt.ContinuousAndCategoricalArray
  ) -> core.DictOf2DArrays:
    """Converts to a single dict of 2d arrays."""
    feat_dict = {}
    cont_ind = 0
    cat_ind = 0
    for converter in self._impl.parameter_converters:
      spec = converter.output_spec
      if spec.type == core.NumpyArraySpecType.CONTINUOUS:
        feat_dict[spec.name] = feat.continuous[:, cont_ind : cont_ind + 1]
        cont_ind += 1
      elif spec.type == core.NumpyArraySpecType.DISCRETE:
        feat_dict[spec.name] = feat.categorical[:, cat_ind : cat_ind + 1]
        cat_ind += 1
      else:
        raise ValueError(
            f'Expected spec to be CONTINUOUS or DISCRETE, saw {spec.type}'
        )
    return core.DictOf2DArrays(feat_dict)

  def to_trials(
      self, feat: vt.ContinuousAndCategoricalArray, labels: np.ndarray
  ) -> Sequence[vz.Trial]:
    # Pass in an empty array to get the structure
    labels_dict = core.DictOf2DArrays(self._impl.to_labels([])).dict_like(
        labels
    )
    return self._impl.to_trials(self._to_dict_2d_arrays(feat), labels_dict)

  def to_parameters(
      self, feat: vt.ContinuousAndCategoricalArray
  ) -> Sequence[vz.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    return self._impl.to_parameters(self._to_dict_2d_arrays(feat))

  # TODO: Move this method to TrialToModelInputConverter, and
  # make this class private to the module.
  @classmethod
  def from_study_config(
      cls,
      study_config: vz.ProblemStatement,
      *,
      scale: bool = True,
      max_discrete_indices: int = 0,
      flip_sign_for_minimization_metrics: bool = True,
      dtype=np.float64,
  ) -> 'TrialToContinuousAndCategoricalConverter':
    """From study config.

    Args:
      study_config:
      scale: If True, scales the parameters to [0, 1] range.
      max_discrete_indices: For DISCRETE and INTEGER parameters that have more
        than this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value. Note this
        default is different from the default in DefaultModelInputConverter.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      dtype: dtype

    Returns:
      TrialToContinuousAndCategoricalConverter
    """

    def create_input_converter(parameter):
      return core.DefaultModelInputConverter(
          parameter,
          scale=scale,
          max_discrete_indices=max_discrete_indices,
          onehot_embed=False,
          float_dtype=dtype,
      )

    def create_output_converter(metric):
      return core.DefaultModelOutputConverter(
          metric,
          flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
          dtype=dtype,
      )

    sc = study_config  # alias, to keep pylint quiet in the next line.
    converter = core.DefaultTrialConverter(
        [create_input_converter(p) for p in sc.search_space.parameters],
        [create_output_converter(m) for m in sc.metric_information],
    )
    return cls(converter)

  @property
  def output_specs(
      self,
  ) -> vt.ContinuousAndCategorical[Sequence[core.NumpyArraySpec]]:
    """Output `NumpyArraySpec`s."""
    continuous = []
    categorical = []
    for converter in self._impl.parameter_converters:
      spec = converter.output_spec
      if spec.type == core.NumpyArraySpecType.CONTINUOUS:
        continuous.append(spec)
      elif spec.type == core.NumpyArraySpecType.DISCRETE:
        categorical.append(spec)
      else:
        raise ValueError(
            f'Expected spec to be CONTINUOUS or DISCRETE, saw {spec.type}'
        )
    return vt.ContinuousAndCategorical(continuous, categorical)

  @property
  def metric_specs(self) -> Sequence[vz.MetricInformation]:
    return [mc.metric_information for mc in self._impl.metric_converters]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for MaskFeatures."""

from absl.testing import parameterized
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax.models import mask_features

from absl.testing import absltest

tfpke = tfp.experimental.psd_kernels


class MaskFeaturesTest(parameterized.TestCase):

  @parameterized.parameters(1, 2, 3, 4)
  def testMaskingAndGradient(self, half_dims: int):
    xs_cont = np.random.randn(10, half_dims)
    xs_cont = np.concatenate(
        [xs_cont, np.full([10, half_dims], np.nan)], axis=-1
    )
    # Generate some integer-encoded categorical data with values between 0 and
    # 5.
    xs_cat = np.broadcast_to(
        np.arange(10)[:, jnp.newaxis] // 2,
        shape=(10, half_dims),
    )
    xs_cat = np.concatenate([xs_cat, np.full([10, half_dims], -1)], axis=-1)
    missing_dim = np.array([False] * half_dims + [True] * half_dims)
    x = tfpke.ContinuousAndCategoricalValues(xs_cont, xs_cat)

    def f(s):
      kernel = tfp.math.psd_kernels.MaternThreeHalves()
      kernel = tfpke.FeatureScaledWithCategorical(
          kernel,
          scale_diag=s,
      )

      kernel = mask_features.MaskFeatures(
          kernel,
          dimension_is_missing=tfpke.ContinuousAndCategoricalValues(
              continuous=missing_dim, categorical=missing_dim
          ),
      )
      return jnp.sum(kernel.matrix(x, x))

    scale_diag = tfpke.ContinuousAndCategoricalValues(
        np.random.uniform(low=1.0, high=10.0, size=(half_dims * 2,)),
        np.random.uniform(low=1.0, high=10.0, size=(half_dims * 2,)),
    )
    value, grad = jax.value_and_grad(f)(scale_diag)
    self.assertTrue(np.all(~np.isnan(value)))
    self.assertTrue(np.all(~np.isnan(grad.continuous)))
    self.assertTrue(np.all(grad.continuous[~missing_dim] != 0))
    self.assertTrue(np.all(grad.continuous[missing_dim] == 0))
    self.assertTrue(np.all(~np.isnan(grad.categorical)))
    self.assertTrue(np.all(grad.categorical[~missing_dim] != 0))
    self.assertTrue(np.all(grad.categorical[missing_dim] == 0))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utility functions for the Pythia protocol code."""

import logging
import threading
from typing import Generic, Optional, Protocol, TypeVar
from vizier import pythia


class _HasErrorDetails(Protocol):
  error_details: str


_T = TypeVar('_T', bound=_HasErrorDetails)


class ResponseWaiter(Generic[_T]):
  """A stored message with a Wait() mechanism.

  This is a bridge between two threads; it stores a message, and
  the reader thread waits until the writer thread calls Report(..., done=True).
  Once Report() is called, it unblocks WaitForRespose().
  """

  def __init__(self):
    self._lock = threading.Lock()
    self._response: Optional[_T] = None
    self._wait = threading.Event()

  def Report(self, update: _T) -> None:
    """Called by the gRPC thread with a message from Vizier.

    When called, this unblocks WaitForResponse().

    Args:
      update:

    Raises:
      PythiaProtocolError
    """
    logging.info('About to take _lock in ResponseWaiter.Report()')
    with self._lock:
      if self._wait.is_set():
        raise pythia.PythiaProtocolError(
            'ResponseWaiter.Report() called after wait is set')
      self._response = update
      self._wait.set()

  def WaitForResponse(self) -> _T:
    """Returns the result or raises an error.

    Raises:
      PythiaProtocolError: Due to a failure of the Pythia protocol.
      VizierDatabaseError: Vizier was unable to service the request.
    """
    self._wait.wait()
    logging.info('About to take _lock in ResponseWaiter.WaitForResponse()')
    with self._lock:
      if self._response is None:
        logging.info('Raise')
        raise pythia.PythiaProtocolError('No response.')
      elif self._response.error_details:
        raise pythia.VizierDatabaseError(self._response.error_details)
      logging.info('No raise -- WaitForResponse done')
      return self._response

  def CancelWait(self) -> None:
    """Cancel the wait."""
    logging.info('About to CancelWait()')
    self._wait.set()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier._src.algorithms.ensemble import ensemble_design

from absl.testing import absltest


class EnsembleDesignTest(absltest.TestCase):

  def testRandom(self):
    indices = [0, 1]
    rewards = [(1, 1), (0, 0), (1, 1)]
    strategy = ensemble_design.RandomEnsembleDesign(indices=indices)

    for reward in rewards:
      strategy.update(reward)
      np.testing.assert_array_equal(strategy.ensemble_probs, [0.5, 0.5])

  def testEXP3IX(self):
    indices = [0, 1]
    rewards = [(1, 1), (0, 0), (1, 1)]
    strategy = ensemble_design.EXP3IXEnsembleDesign(
        indices=indices, stepsize=1.0
    )

    np.testing.assert_array_equal(strategy.ensemble_probs, [0.5, 0.5])

    for reward in rewards:
      strategy.update(reward)

    probs = strategy.ensemble_probs
    self.assertLen(probs, 2)
    self.assertGreater(probs[1], 0.55)
    self.assertGreater(probs[0], 0.3)

  def testEXP3Uniform(self):
    indices = [0, 1]
    rewards = [(1, 1), (0, 0), (1, 1)]
    strategy = ensemble_design.EXP3UniformEnsembleDesign(
        indices=indices, stepsize=1.0
    )

    np.testing.assert_array_equal(strategy.ensemble_probs, [0.5, 0.5])

    for reward in rewards:
      strategy.update(reward)

    probs = strategy.ensemble_probs
    self.assertLen(probs, 2)
    self.assertGreater(probs[1], 0.7)
    self.assertGreater(probs[0], 0.25)

  def testAdaptiveStrategyFixed(self):
    indices = [0, 1]
    rewards = [(1, 1), (0, 0.0), (1, 1)] * 3

    strategy = ensemble_design.AdaptiveEnsembleDesign(
        indices=indices,
        max_lengths=[5, 10],
    )

    np.testing.assert_array_equal(strategy.ensemble_probs, [0.5, 0.5])

    for reward in rewards:
      strategy.update(reward)

    # We should still favor 1 greatly.
    self.assertGreater(strategy.ensemble_probs[1], 0.7)
    self.assertGreater(strategy.observation_probs[1], 0.7)

  def testAdaptiveStrategy(self):
    indices = [0, 1]
    rewards = [(0, 0), (1, 1), (1, 0), (0, 1)]

    strategy = ensemble_design.AdaptiveEnsembleDesign(
        indices=indices,
        max_lengths=[2, 4],
        base_stepsize=1 / np.sqrt(2),
        meta_stepsize=1 / np.sqrt(2),
    )

    np.testing.assert_array_equal(strategy.ensemble_probs, [0.5, 0.5])

    for reward in rewards:
      strategy.update(reward)

    probs = strategy.ensemble_probs
    self.assertLen(probs, 2)
    # We have adapted to favor index 0 quickly.
    self.assertGreater(probs[0], 0.45)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""PSD Kernel for convenience with continuous-only data."""

import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import types

tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


class ContinuousOnly(tfpk.PositiveSemidefiniteKernel):
  """Kernel for using only continuous data and discarding categorical.

  """

  def __init__(
      self,
      kernel: tfpk.PositiveSemidefiniteKernel,
  ):
    parameters = dict(locals())
    self._kernel = kernel

    super(ContinuousOnly, self).__init__(
        feature_ndims=tfpke.ContinuousAndCategoricalValues(
            kernel.feature_ndims, 1),
        dtype=tfpke.ContinuousAndCategoricalValues(
            kernel.dtype, types.INT_DTYPE),
        name='MaskFeatures',
        validate_args=False,
        parameters=parameters,
    )

  @property
  def kernel(self):
    return self._kernel

  def __getattr__(self, name):
    return getattr(self.kernel, name)

  def _apply(
      self, x1: types.ModelInput, x2: types.ModelInput, example_ndims=0
  ) -> jax.Array:
    return self.kernel._apply(x1.continuous, x2.continuous, example_ndims)  # pylint: disable=protected-access

  def _matrix(self, x1: types.ModelInput, x2: types.ModelInput) -> jax.Array:
    return self.kernel._matrix(x1.continuous, x2.continuous)  # pylint: disable=protected-access

  def matrix_over_all_tasks(
      self, x1: types.ModelInput, x2: types.ModelInput
  ) -> jax.Array:
    return self.kernel.matrix_over_all_tasks(x1.continuous, x2.continuous)  # pylint: disable=protected-access

  @classmethod
  def _parameter_properties(cls, dtype):
    return dict(
        kernel=tfp.internal.parameter_properties.BatchedComponentProperties(),
    )


def _continuous_to_cacv(x: jax.Array) -> tfpke.ContinuousAndCategoricalValues:
  return tfpke.ContinuousAndCategoricalValues(
      continuous=x, categorical=jnp.zeros(x.shape[:-1] + (0,))
  )


class EmptyCategoricalKernel(tfpk.PositiveSemidefiniteKernel):
  """Kernel that packs continuous data to ModelInput with empty categorical.

  If `k` is a PSDKernel that operates on `ModelInput` structures, then
  `EmptyCategoricalKernel(k)` is a kernel that operates on arrays of continuous
  data.

  Example:

  ```python
  tfpke = tfp.experimental.psd_kernels
  tfpk = tfp.math.psd_kernels

  # Define a kernel that operates on continuous and categorical values, with the
  # `categorical` field empty.
  kernel = tfpke.FeatureScaledWithCategorical(
      tfpk.ExponentiatedQuadratic(),
      scale_diag=tfpke.ContinuousAndCategoricalValues(
          continuous=jnp.ones([3]),
          categorical=jnp.ones([0])
      )
  )

  # Define an `EmptyCategoricalKernel` that operates on arrays of continuous
  # data.
  empty_cat_kernel = EmptyCategoricalKernel(k)
  xs = np.random.normal([10, 3])
  mat = empty_cat_kernel.matrix(xs, xs)  # Returns a [10, 10] matrix.
  ```
  """

  def __init__(
      self,
      kernel: tfpk.PositiveSemidefiniteKernel,
  ):
    parameters = dict(locals())
    self._kernel = kernel

    super(EmptyCategoricalKernel, self).__init__(
        feature_ndims=kernel.feature_ndims.continuous,
        dtype=kernel.dtype.continuous,
        name='EmptyCategoricalKernel',
        validate_args=False,
        parameters=parameters,
    )

  @property
  def kernel(self):
    return self._kernel

  def __getattr__(self, name):
    return getattr(self.kernel, name)

  def _apply(self, x1: jax.Array, x2: jax.Array, example_ndims: int = 0):
    return self.kernel._apply(  # pylint: disable=protected-access
        _continuous_to_cacv(x1), _continuous_to_cacv(x2), example_ndims
    )

  def _matrix(self, x1: jax.Array, x2: jax.Array):
    return self.kernel._matrix(  # pylint: disable=protected-access
        _continuous_to_cacv(x1), _continuous_to_cacv(x2)
    )

  @classmethod
  def _parameter_properties(cls, dtype):
    return dict(
        kernel=tfp.internal.parameter_properties.BatchedComponentProperties(),
    )


def _flatten(v):
  children = (v.kernel,)
  return (children, None)


def _unflatten(_, children):
  return ContinuousOnly(*children)


jax.tree_util.register_pytree_node(ContinuousOnly, _flatten, _unflatten)
jax.tree_util.register_pytree_node(EmptyCategoricalKernel, _flatten, _unflatten)

tfpke.MultiTaskKernel.register(ContinuousOnly)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.oss.metadata_util."""

from vizier._src.pyvizier.oss import metadata_util
from vizier._src.service import study_pb2

from absl.testing import absltest


class MetadataUtilTest(absltest.TestCase):

  def test_get(self):
    meta_trial = study_pb2.Trial(id='meta_trial')
    trial = study_pb2.Trial(id='trial')
    metadata_util.assign(trial, key='any', ns='', value=meta_trial)
    metadata_util.assign(trial, key='text', ns='x', value='x-value')
    metadata_util.assign(trial, key='text', ns='', value='value')
    metadata_util.assign(trial, key='text', ns='y', value='y-value')
    self.assertEqual(
        metadata_util.get_proto(trial, key='any', ns='', cls=study_pb2.Trial),
        meta_trial)
    self.assertEqual(metadata_util.get(trial, key='text', ns=''), 'value')
    self.assertEqual(metadata_util.get(trial, key='text', ns='x'), 'x-value')

    self.assertIsNone(
        metadata_util.get_proto(trial, key='any', ns='', cls=study_pb2.Study))
    self.assertIsNone(metadata_util.get(trial, key='TYPO', ns=''))
    self.assertIsNone(
        metadata_util.get_proto(trial, key='TYPO', ns='', cls=study_pb2.Trial))

  def test_assign(self):
    trial = study_pb2.Trial(id='trial')
    metadata_util.assign(trial, key='k', ns='', value='value')
    self.assertEqual(metadata_util.get(trial, key='k', ns=''), 'value')
    metadata_util.assign(
        trial, key='k', ns='', value='222', mode='insert_or_assign'
    )
    self.assertEqual(metadata_util.get(trial, key='k', ns=''), '222')


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter that shifts the values of the parameters in Evaluation."""

import copy
from typing import Sequence

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


class ShiftingExperimenter(experimenter.Experimenter):
  """ShiftingExperimenter shifts the parameters of suggestions in Evaluate."""

  def __init__(
      self,
      exptr: experimenter.Experimenter,
      shift: np.ndarray,
      should_restrict=True,
  ):
    """ShiftingExperiment shifts parameter values before passing to exptr.

    Currently only supports flat double search spaces. Note that when
    should_restrict is True, the parameter bounds of the search space are
    RESTRICTED to never cause a parameter to
    exceed the underlying experimenter's parameter bounds, so the problem
    statement will change.

    Args:
      exptr: Underlying experimenter to be wrapped.
      shift: Shift that broadcasts to array of shape (dimension,).
      should_restrict: Whether to restrict the parameter bounds of search space.

    Raises:
      ValueError: Non-positive dimension or non-broadcastable/large shift.
    """
    self._exptr = exptr
    exptr_problem_statement = exptr.problem_statement()

    if exptr_problem_statement.search_space.is_conditional:
      raise ValueError('Search space should not have conditional'
                       f' parameters  {exptr_problem_statement}')
    dimension = len(exptr_problem_statement.search_space.parameters)
    if dimension <= 0:
      raise ValueError(f'Invalid dimension: {dimension}')
    try:
      # Attempts a broadcast to check broadcasting.
      self._shift = np.broadcast_to(shift, (dimension,))
    except ValueError as broadcast_err:
      raise ValueError(
          f'Shift {shift} is not broadcastable for dim: {dimension}.'
          '\n') from broadcast_err

    # Converter should be in the underlying extpr space.
    self._converter = converters.TrialToArrayConverter.from_study_config(
        study_config=exptr_problem_statement, scale=False)

    self._problem_statement = copy.deepcopy(exptr_problem_statement)
    if should_restrict:
      self._problem_statement.search_space = pyvizier.SearchSpace()

      for parameter, shift in zip(
          exptr_problem_statement.search_space.parameters, self._shift
      ):
        if parameter.type != pyvizier.ParameterType.DOUBLE:
          raise ValueError(f'Non-double parameters {parameter}')
        if (bounds := parameter.bounds) is not None:
          if abs(shift) >= bounds[1] - bounds[0]:
            raise ValueError(
                f'Bounds {bounds} may need to be extended'
                f'as shift {shift} is too large '
            )
          # Shift the bounds to maintain valid bounds.
          if shift >= 0:
            new_bounds = (bounds[0], bounds[1] - shift)
          else:
            # Shift is negative so this restricts the bounds.
            new_bounds = (bounds[0] - shift, bounds[1])
          self._problem_statement.search_space.add(
              pyvizier.ParameterConfig.factory(
                  name=parameter.name,
                  bounds=new_bounds,
                  scale_type=parameter.scale_type,
                  default_value=parameter.default_value,
                  external_type=parameter.external_type,
              )
          )

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]) -> None:
    """Evaluate the trials after shifting their parameters by +shift."""
    self._offset(suggestions, self._shift)
    self._exptr.evaluate(suggestions)
    self._offset(suggestions, -self._shift)

  def _offset(self, suggestions: Sequence[pyvizier.Trial],
              shift: np.ndarray) -> None:
    """Offsets the suggestions parameter values in place."""
    for suggestion in suggestions:
      features = self._converter.to_features([suggestion])
      new_parameters = self._converter.to_parameters(features + shift)[0]
      suggestion.parameters = new_parameters

  def __repr__(self):
    return f'ShiftingExperimenter({self._shift}) on {str(self._exptr)}'



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Types library for vizier/_src/jax."""

from typing import Any, Generic, Iterable, Mapping, Optional, Sequence, TypeVar, Union
import equinox as eqx
from flax import struct
from flax.core import scope as flax_scope
import jax
from jax import numpy as jnp
from jax.typing import ArrayLike
import jaxtyping as jt
import numpy as np

# Integer dtype for categorical data.
INT_DTYPE = np.int32


# We define our own Array type since `jax.typing.Array` and `chex.Array` both
# include scalar types, which result in type errors when array
# methods/properties like `.shape` are accessed.
Array = Union[np.ndarray, jax.Array]


# TODO: Add a `concatenate` method.
class PaddedArray(eqx.Module):
  """Padded Array as a pytree.

  There is no validation done in `__init__`. In order to get a validated
  instance, Use a converter object or a factory method outside the jit scope.

  Attributes:
    padded_array:
    _original_shape: 1D array of length equal to `padded_array.ndims`. The k-th
      element must be less than or equal to `padded_array.shape[k]`.
    _nopadding_done: Must be set to True iff `_original_shape` is the same as
      padded_array.shape.
    _mask: Same shape as padded_array. True where padded_array's value is the
      original array value as opposed to the fill value.
  """

  # TODO: Rename "padded_array" to a shorter name like "padded".
  padded_array: jt.Shaped[jax.Array, '...'] = eqx.field(converter=jnp.asarray)
  fill_value: Any = eqx.field(converter=jnp.asarray)
  # TODO: Make `_original_shape` public.
  _original_shape: jt.Int[jax.Array, '_'] = eqx.field(converter=jnp.asarray)

  # TODO: Make _mask an inferred property. The current
  # implementation is not memory efficient.
  _mask: jt.Bool[jax.Array, '...'] = eqx.field(converter=jnp.asarray)
  _nopadding_done: bool = eqx.field(static=True, default=False, kw_only=True)

  @property
  def shape(self) -> tuple[int, ...]:
    """Returns the shape of the padded array."""
    return self.padded_array.shape

  def replace_array(self, array: jt.Shaped[jax.Array, '...']) -> 'PaddedArray':
    """Replaces the original array values, maintaining the padding."""
    return PaddedArray.from_array(array, self.shape, fill_value=self.fill_value)

  def replace_fill_value(self, fill_value: Any) -> 'PaddedArray':
    """Replaces the padded fill values."""
    # TODO: Consider optimizing when fill_value == self.fill_value.
    if self._nopadding_done:
      return self
    else:
      return PaddedArray(
          jnp.where(self._mask, self.padded_array, fill_value),
          fill_value=fill_value,
          _original_shape=self._original_shape,
          _mask=self._mask,
          _nopadding_done=self._nopadding_done,
      )

  @classmethod
  def as_padded(cls, array: jt.Shaped[jax.Array, '...']) -> 'PaddedArray':
    """Converts a regular array as PaddedArray type, with no actual padding.

    NOTE This is implemented as a separate method instead of setting default
    for `fill_value=np.nan` in `from_array` method. `jnp.pad`` automatically
    casts `nan` to `0` for integer arrays, which can cause unexpected
    behavior, and we want people to always explicitly set the fill value
    unless they know for sure that padding would not occur.

    Args:
      array:

    Returns:
      PaddedArray.
    """
    return PaddedArray.from_array(array, array.shape, fill_value=np.nan)

  @classmethod
  def from_array(
      cls,
      array: jt.Shaped[jax.Array, '...'],
      target_shape: Sequence[int],
      *,
      fill_value: Any
  ) -> 'PaddedArray':
    """Factory function to guarantee a self-consistent creation."""
    spec = [(0, dnew - d) for d, dnew in zip(array.shape, target_shape)]
    mask_array = jnp.pad(
        jnp.ones_like(array, dtype=bool), spec, constant_values=False
    )
    new_array = jnp.pad(array, spec, constant_values=fill_value)
    return PaddedArray(
        padded_array=new_array,
        fill_value=fill_value,
        _original_shape=array.shape,
        _mask=mask_array,
        _nopadding_done=array.shape == target_shape,
    )

  @property
  def is_missing(self) -> tuple[jt.Bool[jax.Array, '...']]:
    """Mask per dimension padded. Arrays have shape [N1], [N2], ..., [Nk]."""
    return tuple(
        jnp.arange(s1) >= s2
        for s1, s2 in zip(self.padded_array.shape, self._original_shape)
    )

  def unpad(self) -> jt.Shaped[jax.Array, '...']:
    """Can be used in jit scope only if original shape == padded shape."""
    if self._nopadding_done:
      return self.padded_array
    return jax.lax.slice(
        self.padded_array,
        [0] * self.padded_array.ndim,
        self._original_shape.tolist(),
    )


ArrayTree = Union[ArrayLike, Iterable['ArrayTree'], Mapping[Any, 'ArrayTree']]

# An ArrayTree that allows None values.
ArrayTreeOptional = Union[
    ArrayLike,
    Iterable[Optional['ArrayTreeOptional']],
    Mapping[Any, Optional['ArrayTreeOptional']],
]
ParameterDict = flax_scope.Collection
ModelState = flax_scope.VariableDict


_T = TypeVar('_T')


@struct.dataclass
class ContinuousAndCategorical(Generic[_T]):
  continuous: _T
  categorical: _T


# TODO: Make it a private type inside jnp_converters.py.
ContinuousAndCategoricalArray = ContinuousAndCategorical[Array]

ModelInput = ContinuousAndCategorical[PaddedArray]


class ModelData(eqx.Module):
  features: ModelInput
  labels: PaddedArray


# Tuple representing a box constraint of the form (lower, upper) bounds.
Bounds = tuple[Optional[ArrayTreeOptional], Optional[ArrayTreeOptional]]


# TODO: Deprecate it in favor of
# PrecomputedPredictive for full predictive state including cholesky, and
# StochasticProcessWithCoroutine for computing log probs.
@struct.dataclass
class GPState:
  """State that changes at each iteration."""

  data: ModelData
  model_state: ModelState



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""High-level wrappers for stochastic process hyperparameter optimizers."""

import datetime
import time
from typing import Any, Optional

from absl import logging
import attr
import chex
import equinox as eqx
from flax import struct
import jax
from jax import numpy as jnp
import jaxopt
import tree
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax.optimizers import core


@struct.dataclass
class LbfgsBOptions:
  """L-BFGS-B is a version of L-BFGS that incorporates box constraints.

  https://digital.library.unt.edu/ark:/67531/metadc666315/m2/1/high_res_d/204262.pdf

  Attributes:
    num_line_search_steps: Maximum number of line search steps.
    random_restarts: Must be positive; number of random initializations for the
      optimization.
    tol: Tolerance for stopping criteria.
    maxiter: Max number of iterations.
    best_n: Number of best values to return from the initializations; must be
      less than or equal to `random_restarts`.
  """

  num_line_search_steps: int = struct.field(kw_only=True, default=20)
  # TODO: Remove this since now it's a part of optimizer API.
  random_restarts: Optional[int] = struct.field(
      kw_only=True, default=None, pytree_node=False
  )
  tol: float = struct.field(kw_only=True, default=1e-8)
  maxiter: int = struct.field(kw_only=True, default=50)
  # TODO: Remove best_n. Now it's a part of optimizer API.
  best_n: Optional[int] = struct.field(
      kw_only=True, default=None, pytree_node=False
  )


def _is_leaf(x: float) -> bool:
  return jax.tree_util.treedef_is_leaf(jax.tree_util.tree_structure(x))


def _none_to_inf(b: float, inf: float, params: chex.ArrayTree):
  """Converts None bounds to inf or -inf to pass to the optimizer."""
  if b is None:
    b = inf
  if _is_leaf(b):
    # Broadcast scalars to the parameter structure.
    return tree.map_structure(lambda x: b * jnp.ones_like(x), params)
  else:
    # For structured bounds, replace `None`s in the structure with `inf`.
    return tree.map_structure(
        lambda b_, x: jnp.ones_like(x) * (inf if b_ is None else b_), b, params
    )


# TODO: Remove support for broadcasting array -> arraytree
# and reduce maintenance burden. Alternatively, move to a separate file
# so other libraries can use it too.
def _get_bounds(
    params: core.Params,
    constraints: Optional[sp.Constraint],
) -> Optional[tuple[chex.ArrayTree, chex.ArrayTree]]:
  """Returns (Lower, upper) ArrayTrees with the same shape as params."""
  if constraints is None:
    return None
  else:
    lb = _none_to_inf(constraints.bounds[0], -jnp.inf, params)
    ub = _none_to_inf(constraints.bounds[1], jnp.inf, params)
    logging.info(
        'constraints\n: %s converted to bounds:\n %s', constraints, (lb, ub)
    )
    return (lb, ub)


def _unbatch_params(batched_params: core.Params) -> list[core.Params]:
  batch_size = jax.tree_util.tree_leaves(batched_params)[0].shape[0]
  return list(
      map(
          lambda i: jax.tree_map(lambda p: p[i], batched_params),
          range(batch_size),
      )
  )


@attr.define
class JaxoptScipyLbfgsB(core.Optimizer[core.Params]):
  """Jaxopt's wrapper for scipy L-BFGS-B optimizer."""

  _options: LbfgsBOptions = attr.field(default=LbfgsBOptions())
  _speed_test: bool = attr.field(kw_only=True, default=False)
  _max_duration: Optional[datetime.timedelta] = None

  def __call__(
      self,
      init_params: core.Params,
      loss_fn: core.LossFunction[core.Params],
      rng: jax.Array,
      *,
      constraints: Optional[sp.Constraint] = None,
      best_n: Optional[int] = None,
  ) -> tuple[core.Params, chex.ArrayTree]:
    # L-BFGS-B is deterministic given initial parameters.
    del rng

    # L-BFGS-B may be used on unconstrained problems (in which case it is
    # slightly different from L-BFGS, in that it uses the Cauchy point/subspace
    # minimization to choose the line search direction). Bounds must be None or
    # a tuple of size 2. The tuple must contain lower/upper bounds, which may be
    # None or a pytree with the same structure as the model parameters returned
    # by setup (otherwise Jaxopt will raise an error).
    jax.monitoring.record_event('/vizier/jax/optimizers/scipylbfgsb/called')
    lbfgsb = jaxopt.ScipyBoundedMinimize(
        fun=loss_fn,
        method='L-BFGS-B',
        maxiter=self._options.maxiter,
        options={
            'gtol': self._options.tol,
            'maxls': self._options.num_line_search_steps,
        },
        jit=False,
        has_aux=True,
        value_and_grad=eqx.filter_jit(
            eqx.filter_value_and_grad(loss_fn, has_aux=True)
        ),
    )
    train_times = []
    losses = []
    params = []
    metrics = {}

    init_params = _unbatch_params(init_params)
    bounds = _get_bounds(init_params[0], constraints)

    logging.info(
        'Using SCIPY L-BFGS-B w/ %d initializations: %s',
        len(init_params),
        init_params,
    )
    for p in init_params:
      if self._max_duration is not None and train_times:
        expected_worst_case_duration = max(train_times) + sum(train_times)
        if expected_worst_case_duration >= self._max_duration.total_seconds():
          logging.warn(
              'Maximum duration of %s is expected to be surpassed.'
              ' Completed only %s initializations out of %s.',
              self._max_duration,
              len(params),
              len(init_params),
          )
          break
      start_time = time.time()
      position, opt_state = lbfgsb.run(init_params=p, bounds=bounds)
      train_times.append(time.time() - start_time)
      losses.append(opt_state.fun_val)
      params.append(position)
      logging.info(
          'Loss: %s, last step time: %s, ScipyMinimizeInfo: %s',
          opt_state.fun_val,
          train_times[-1],
          opt_state,
      )
    losses = jnp.asarray(losses)
    all_params = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *params)

    metrics['loss'] = losses[jnp.newaxis, :]
    if self._speed_test:
      metrics['train_time'] = train_times
    return (
        core.get_best_params(losses, all_params, best_n=best_n),
        metrics,
    )

  # TODO: Remove dependencies on this property.
  @property
  def best_n(self) -> int:
    return self._options.best_n or 1


def _run_parallel_lbfgs(
    loss_fn: core.LossFunction[core.Params],
    init_params_batch: core.Params,
    *,
    bounds: Optional[tuple[chex.ArrayTree, chex.ArrayTree]],
    options: LbfgsBOptions,
) -> tuple[core.Params, Any]:
  """Called by JaxoptLbfgsB."""

  def _run_one_lbfgs(
      init_params: core.Params,
  ) -> tuple[core.Params, Any]:
    lbfgsb = jaxopt.LBFGSB(
        fun=loss_fn,
        maxls=options.num_line_search_steps,
        tol=options.tol,
        maxiter=options.maxiter,
        has_aux=True,
    )
    return lbfgsb.run(init_params=init_params, bounds=bounds)

  # We chose map over vmap because some of the lbfgs runs may terminate early.
  # pmap is also not fit for our use case, because we typically have a single
  # processor.
  return jax.lax.map(_run_one_lbfgs, init_params_batch)


@attr.define
class JaxoptLbfgsB(core.Optimizer[core.Params]):
  """Jaxopt's L-BFGS-B optimizer.

  Jaxopt calls the Scipy or Jax implementation of L-BFGS-B.
  L-BFGS-B is a version of L-BFGS that incorporates box constraints on
  parameters.
  https://digital.library.unt.edu/ark:/67531/metadc666315/m2/1/high_res_d/204262.pdf

  Attributes:
    num_line_search_steps: Maximum number of line search steps.
    random_restarts: Must be positive; number of random initializations for the
      optimization.
    best_n: Number of best values to return from the initializations; must be
      less than or equal to `random_restarts`.
     use_scipy: Uses the scipy version of L-BFGS-B. If False, uses the pure JAX
       L-BFGS-B in Jaxopt, which runs on accelerators.
     tol: Tolerance for stopping criteria.
     maxiter: Max number of iterations.
    _speed_test: If True, return speed test results.
  """

  _options: LbfgsBOptions = attr.field(default=LbfgsBOptions())
  _speed_test: bool = attr.field(kw_only=True, default=False)

  def __call__(
      self,
      init_params: core.Params,
      loss_fn: core.LossFunction[core.Params],
      rng: jax.Array,
      *,
      constraints: Optional[sp.Constraint] = None,
      best_n: Optional[int] = None,
  ) -> tuple[core.Params, chex.ArrayTree]:
    metrics = {}
    logging.info('Using JAX L-BFGS-B w/ RNG: %s', rng)
    start_time = time.time()
    bounds = _get_bounds(_unbatch_params(init_params)[0], constraints)
    lbfgsb = jaxopt.LBFGSB(
        fun=eqx.filter_jit(loss_fn),
        value_and_grad=eqx.filter_value_and_grad(loss_fn, has_aux=True),
        maxls=self._options.num_line_search_steps,
        tol=self._options.tol,
        maxiter=self._options.maxiter,
        has_aux=True,
        jit=True,
        unroll=True,
    )

    all_params, opt_states = jax.lax.map(
        lambda p: lbfgsb.run(init_params=p, bounds=bounds),
        init_params,
    )  # pylint: disable=g-long-lambda

    losses = jnp.asarray(opt_states.value)
    metrics['train_time'] = time.time() - start_time
    metrics['loss'] = losses[jnp.newaxis, :]
    return (
        core.get_best_params(losses, all_params, best_n=best_n),
        metrics,
    )

  # TODO: Remove dependencies on this property.
  @property
  def best_n(self) -> int:
    return self._options.best_n or 1

  @property
  def jittable(self) -> bool:
    return True



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Atari100k benchmarks from Dopamine.

Reference on benchmark noise and settings:

`Deep RL at the Edge of the Statistical Precipice.`
(https://arxiv.org/abs/2108.13264)

"""
# pylint:disable=dangerous-default-value
from typing import Dict, Optional, Sequence, Union

from absl import logging
from dopamine.discrete_domains import iteration_statistics
from dopamine.labs.atari_100k import atari_100k_rainbow_agent
from dopamine.labs.atari_100k import eval_run_experiment
import gin

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter


def create_agent_fn(
    sess,  # pylint: disable=unused-argument
    environment,
    seed: Optional[int] = None,
    summary_writer=None) -> atari_100k_rainbow_agent.Atari100kRainbowAgent:
  """Helper function for creating full rainbow-based Atari 100k agent."""
  return atari_100k_rainbow_agent.Atari100kRainbowAgent(
      num_actions=environment.action_space.n,
      seed=seed,
      summary_writer=summary_writer)


@gin.configurable
class WrappedRunner(eval_run_experiment.MaxEpisodeEvalRunner):
  """Wraps the original Dopamine Runner for Vizier's convenience."""

  def __init__(self, base_dir: str = '/tmp/'):
    super().__init__(base_dir=base_dir, create_agent_fn=create_agent_fn)

  def run_trial(self) -> iteration_statistics.IterationStatistics:
    """Replacement of `run_experiment()` to avoid TensorBoard summary logging."""

    statistics = iteration_statistics.IterationStatistics()

    if self._num_iterations <= self._start_iteration:
      logging.warning('num_iterations (%d) < start_iteration(%d)',
                      self._num_iterations, self._start_iteration)
      return statistics

    for _ in range(self._start_iteration, self._num_iterations):
      self._run_train_phase(statistics)
      self._run_eval_phase(statistics)

    return statistics


GinParameterType = Union[float, int, str]


def default_search_space() -> pyvizier.SearchSpace:
  """Produces a reasonable SearchSpace for tuning the Rainbow training process."""
  ss = pyvizier.SearchSpace()
  ss.root.add_float_param(
      'JaxDQNAgent.gamma',
      0.7,
      0.999999,
      scale_type=pyvizier.ScaleType.REVERSE_LOG)
  ss.root.add_int_param('JaxDQNAgent.update_horizon', 1, 20)
  ss.root.add_int_param('JaxDQNAgent.update_period', 1, 10)
  ss.root.add_int_param('JaxDQNAgent.target_update_period', 1, 10000)
  ss.root.add_int_param('JaxDQNAgent.min_replay_history', 100, 100000)
  ss.root.add_float_param(
      'JaxDQNAgent.epsilon_train',
      0.0000001,
      1.0,
      scale_type=pyvizier.ScaleType.LOG)
  ss.root.add_int_param('JaxDQNAgent.epsilon_decay_period', 1000, 10000)
  ss.root.add_bool_param('JaxFullRainbowAgent.noisy')
  ss.root.add_bool_param('JaxFullRainbowAgent.dueling')
  ss.root.add_bool_param('JaxFullRainbowAgent.double_dqn')
  ss.root.add_int_param('JaxFullRainbowAgent.num_atoms', 1, 100)
  ss.root.add_bool_param('Atari100kRainbowAgent.data_augmentation')
  ss.root.add_float_param(
      'create_optimizer.learning_rate',
      0.0000001,
      1.0,
      scale_type=pyvizier.ScaleType.LOG)
  ss.root.add_float_param(
      'create_optimizer.eps', 0.0000001, 1.0, scale_type=pyvizier.ScaleType.LOG)

  return ss


class Atari100kExperimenter(experimenter.Experimenter):
  """Atari100k Experimenter."""

  def __init__(self,
               game_name: str = 'Pong',
               agent_name: str = 'DER',
               initial_gin_bindings: Dict[str, GinParameterType] = {}):
    """Initializes the Atari100k Experimenter.

    Args:
      game_name: Atari game name. Can be one of 57 games.
      agent_name: Name of the base config file to use. Note that the
        corresponding gin files contain reported/default values. To run them,
        just send in a trivial study_config, which leads to a separate XM run
        containing baseline results.
      initial_gin_bindings: Initial gin bindings to use. Useful for making tests
        small. Will be overridden by the trial's gin bindings.
    """
    self._game_name = game_name
    assert agent_name in ['DER', 'DrQ', 'DrQ_eps', 'OTRainbow']
    self._gin_file = f'vizier/_src/benchmarks/experimenters/atari100k_configs/{agent_name}.gin'
    self._initial_gin_bindings = initial_gin_bindings

  def problem_statement(self) -> pyvizier.ProblemStatement:
    ss = default_search_space()
    problem_statement = pyvizier.ProblemStatement(search_space=ss)
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='eval_average_return',
            goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))
    return problem_statement

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for trial in suggestions:
      with gin.unlock_config():
        # Lock in initial values.
        gin.parse_config_file(self._gin_file)
        gin.bind_parameter('atari_lib.create_atari_environment.game_name',
                           self._game_name)
        for parameter_name in self._initial_gin_bindings:
          gin.bind_parameter(parameter_name,
                             self._initial_gin_bindings[parameter_name])

        # Lock in trial parameters.
        for parameter_name in trial.parameters:
          gin.bind_parameter(parameter_name,
                             trial.parameters[parameter_name].value)

      # Run actual training + inference.
      runner = WrappedRunner()
      statistics = runner.run_trial()
      logging.info('Statistics: %s', statistics.data_lists)

      # Add Intermediate Measurements.
      num_intermediate_measurements = len(
          statistics.data_lists['eval_average_return'])
      for i in range(num_intermediate_measurements):
        measurement = pyvizier.Measurement()
        measurement.metrics['train_average_return'] = statistics.data_lists[
            'train_average_return'][i]
        measurement.metrics[
            'train_average_steps_per_second'] = statistics.data_lists[
                'train_average_steps_per_second'][i]
        measurement.metrics['eval_average_return'] = statistics.data_lists[
            'eval_average_return'][i]
        trial.measurements.append(measurement)

      # Final Measurement.
      trial.complete(trial.measurements[-1])



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Generic interfaces and template for evolutionary algorithms.

The evolutionary algorithm and Vizier interaction workflow is:
  * Algorithm generates `Offsprings`, which contains information about the
  "genes". `PopulationConverter` converts them to `TrialSuggestions`.
  * User evaluates `TrialSuggestions` and obtains completed `Trial` objects.
  `PopulationConverter` converts them to `Population`, which is what the
  algorithm receives.
  * This template does not dictate the nature of these conversions. They can be
  stochastic, deterministic, bijective, injective, inverse of each other or not.

Why are Offsprings and Population serializable?
  * `Offsprings` is serializable so that the original genes can be preserved
  through the conversions.
  * `Population` is serializable so that the algorithm can serialize its state.

A canonical evolutionary algorithm consists of the following operations:
  * `Sampler` generates fresh `Offsprings` independent of `Population`.
    It is generally used for sampling an initial batch of `Offsprings`.
  * `Survival` selects a subset of `Population` to keep.
  * `Mutation` mutates `Population` to generate new `Offsprings`.

To use the provided template, one should implement a subclass of `Offsprings`,
`Population`, `PopulationConverter`, `Sampler`, `Survival`, and `Mutation`.
"""

import abc
from typing import Generic, Optional, Sequence, TypeVar, Union

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_OffspringsType = TypeVar('_OffspringsType', bound=serializable.Serializable)


class Sampler(Generic[_OffspringsType], abc.ABC):
  """Creates new offsprings, typically at the start of evolutionary strategy."""

  @abc.abstractmethod
  def sample(self, count: int) -> _OffspringsType:
    """Generate offsprings."""


# Temporary typevar, used for defininig Population interface.
_P = TypeVar('_P', bound=serializable.Serializable)


class Population(serializable.Serializable, abc.ABC):
  """Typically contains genes (x-values) and scores (y-values).

  The supported operations are similar to python sequence, except that slicing
  always results in a `Population` of length 1.
  """

  @abc.abstractmethod
  def __len__(self) -> int:
    pass

  @abc.abstractmethod
  def __getitem__(
      self: _P,
      index: Union[int, slice],
     
  ) -> _P:
    pass

  @abc.abstractmethod
  def __add__(self: _P, other: _P) -> _P:
    pass


# Actual Typevar. Covariant with `Population`.
_PopulationType = TypeVar('_PopulationType')


class PopulationConverter(abc.ABC, Generic[_PopulationType, _OffspringsType]):

  @abc.abstractmethod
  def to_population(self,
                    completed: Sequence[vz.CompletedTrial]) -> _PopulationType:
    """Adds trials to the population."""

  @abc.abstractmethod
  def to_suggestions(
      self, offsprings: _OffspringsType) -> Sequence[vz.TrialSuggestion]:
    """Converts offsprings to suggestions."""


class Survival(abc.ABC, Generic[_PopulationType]):

  @abc.abstractmethod
  def select(self, population: _PopulationType) -> _PopulationType:
    """Survival mechanism."""


class Mutation(abc.ABC, Generic[_PopulationType, _OffspringsType]):

  @abc.abstractmethod
  def mutate(self, population: _PopulationType, count: int) -> _OffspringsType:
    """Generate offsprings, whose contain count many offsprings."""


class CanonicalEvolutionDesigner(vza.PartiallySerializableDesigner,
                                 Generic[_PopulationType, _OffspringsType]):
  """Evolution algorithm template."""

  def __init__(
      self,
      converter: PopulationConverter[_PopulationType, _OffspringsType],
      sampler: Sampler[_OffspringsType],
      survival: Survival[_PopulationType],
      adaptation: Mutation[_PopulationType, _OffspringsType],
      *,
      initial_population: Optional[_PopulationType] = None,
      first_survival_after: Optional[int] = None,
      population_size: int = 50,
  ):
    """Init.

    Args:
      converter:
      sampler:
      survival:
      adaptation:
      initial_population: The initial population to seed the evolution.
      first_survival_after: Apply the survival step after observing this many
        trials. If unset, it defaults to twice the `population_size`.
      population_size: Survival steps reduce the population to this size.
    """
    self._survival = survival
    self._adaptation = adaptation
    self._converter = converter
    self._sampler = sampler
    self._population_size = population_size
    self._first_survival_after = (
        first_survival_after or self._population_size * 2
    )
    self._num_trials_seen = 0
    self._population = initial_population or converter.to_population([])

  @property
  def converter(self) -> PopulationConverter[_PopulationType, _OffspringsType]:
    return self._converter

  @property
  def population(self) -> _PopulationType:
    return self._population

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    count = count or self._population_size
    if self._num_trials_seen < self._first_survival_after:
      return self._converter.to_suggestions(self._sampler.sample(count))
    return self._converter.to_suggestions(
        self._adaptation.mutate(self._population, count))

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    completed = completed.trials
    self._num_trials_seen += len(completed)
    candidates = self._population + self._converter.to_population(completed)
    self._population = self._survival.select(candidates)

  def load(self, metadata: vz.Metadata):
    self._population = type(self._population).recover(metadata)

  def dump(self) -> vz.Metadata:
    return self._population.dump()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from jax import numpy as jnp
from vizier._src.algorithms.designers import scalarization
from absl.testing import absltest


class ScalarizationsTest(absltest.TestCase):

  def test_linear_scalarizer(self):
    scalarizer = scalarization.LinearScalarization(
        weights=jnp.array([0.1, 0.2])
    )
    self.assertAlmostEqual(scalarizer(jnp.array([3.0, 4.5])), 1.2)

  def test_chebyshev_scalarizer(self):
    scalarizer = scalarization.ChebyshevScalarization(
        weights=jnp.array([0.1, 0.2])
    )
    self.assertAlmostEqual(scalarizer(jnp.array([3.0, 4.5])), 0.3)

  def test_hypervolume_scalarizer(self):
    scalarizer = scalarization.HyperVolumeScalarization(
        weights=jnp.array([0.1, 0.2])
    )
    self.assertAlmostEqual(scalarizer(jnp.array([3.0, 4.5])), 22.5)

  def test_hypervolume_scalarizer_with_reference(self):
    scalarizer = scalarization.HyperVolumeScalarization(
        weights=jnp.array([0.1, 0.2]), reference_point=jnp.array([-1])
    )
    self.assertAlmostEqual(scalarizer(jnp.array([3.0, 4.5])), 27.5)

  def test_augmented_scalarizer(self):
    scalarizer = scalarization.LinearAugmentedScalarization(
        weights=jnp.array([0.1, 0.2]),
        scalarization_factory=scalarization.HyperVolumeScalarization,
    )
    # Should be the sum of hypervolume and linear scalarizations.
    self.assertAlmostEqual(scalarizer(jnp.array([3.0, 4.5])), 23.7)

if __name__ == "__main__":
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utility functions for creating GRPC stubs."""

import functools
from typing import Optional
from absl import logging
import grpc

from vizier._src.service import pythia_service_pb2_grpc
from vizier._src.service import vizier_service_pb2_grpc


def _create_channel(
    endpoint: str, timeout: Optional[float] = None
) -> grpc.Channel:
  """Creates GRPC channel."""
  logging.info('Securing channel to %s.', endpoint)
  channel = grpc.insecure_channel(endpoint)
  grpc.channel_ready_future(channel).result(timeout=timeout)
  logging.info('Created channel to %s.', endpoint)
  return channel


@functools.lru_cache(maxsize=128)
def create_pythia_server_stub(
    endpoint: str, timeout: Optional[float] = 10.0
) -> pythia_service_pb2_grpc.PythiaServiceStub:
  """Creates the GRPC stub.

  This method uses LRU cache so we create a single stub per endpoint (which is
  effectively one per binary). Stub and channel are both thread-safe and can
  take a while to create. The LRU cache makes binaries run faster, especially
  for unit tests.

  Args:
    endpoint: Pythia server endpoint.
    timeout: Timeout in seconds. If None, no timeout will be used.

  Returns:
    Pythia Server stub at endpoint.
  """
  return pythia_service_pb2_grpc.PythiaServiceStub(
      _create_channel(endpoint, timeout)
  )


@functools.lru_cache(maxsize=128)
def create_vizier_server_stub(
    endpoint: str, timeout: Optional[float] = 10.0
) -> vizier_service_pb2_grpc.VizierServiceStub:
  """Creates the GRPC stub.

  This method uses LRU cache so we create a single stub per endpoint (which is
  effectively one per binary). Stub and channel are both thread-safe and can
  take a while to create. The LRU cache makes binaries run faster, especially
  for unit tests.

  Args:
    endpoint: Vizier server endpoint.
    timeout: Timeout in seconds. If None, no timeout will be used.

  Returns:
    Vizier Server stub at endpoint.
  """
  return vizier_service_pb2_grpc.VizierServiceStub(
      _create_channel(endpoint, timeout)
  )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gp_bandit."""

from typing import Callable, Union
import unittest
from unittest import mock

import jax
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_bandit
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.algorithms.optimizers import eagle_strategy as es
from vizier._src.algorithms.optimizers import lbfgsb_optimizer as lo
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.algorithms.testing import test_runners
from vizier._src.jax import types
from vizier.jax import optimizers
from vizier.pyvizier import converters
from vizier.pyvizier.converters import padding
from vizier.testing import test_studies
from vizier.utils import profiler

from absl.testing import absltest
from absl.testing import parameterized


ard_optimizer = optimizers.default_optimizer()
vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
    strategy_factory=es.VectorizedEagleStrategyFactory(),
    max_evaluations=10,
)
lbfgsb_optimizer_factory = lo.LBFGSBOptimizerFactory()


def _build_mock_continuous_array_specs(n):
  continuous_spec = mock.create_autospec(converters.NumpyArraySpec)
  continuous_spec.type = converters.NumpyArraySpecType.CONTINUOUS
  continuous_spec.num_dimensions = 1
  return [continuous_spec] * n


def _setup_lambda_search(
    f: Callable[[float], float], num_trials: int = 100
) -> tuple[gp_bandit.VizierGPBandit, list[vz.Trial], vz.ProblemStatement]:
  """Sets up a GP designer and outputs completed studies for `f`.

  Args:
    f: 1D objective to be optimized, i.e. f(x), where x is a scalar in [-5., 5.)
    num_trials: Number of mock "evaluated" trials to return.

  Returns:
  A GP designer set up for the problem of optimizing the objective, without any
  data updated.
  Evaluated trials against `f`.
  """
  assert (
      num_trials > 0
  ), f'Must provide a positive number of trials. Got {num_trials}.'

  search_space = vz.SearchSpace()
  search_space.root.add_float_param('x0', -5.0, 5.0)
  problem = vz.ProblemStatement(
      search_space=search_space,
      metric_information=vz.MetricsConfig(
          metrics=[
              vz.MetricInformation('obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
          ]
      ),
  )

  suggestions = quasi_random.QuasiRandomDesigner(
      problem.search_space, seed=1
  ).suggest(num_trials)

  obs_trials = []
  for idx, suggestion in enumerate(suggestions):
    trial = suggestion.to_trial(idx)
    x = suggestions[idx].parameters['x0'].value
    trial.complete(vz.Measurement(metrics={'obj': f(x)}))
    obs_trials.append(trial)

  gp_designer = gp_bandit.VizierGPBandit(problem, ard_optimizer=ard_optimizer)
  return gp_designer, obs_trials, problem


def _compute_mse(
    designer: gp_bandit.VizierGPBandit,
    test_trials: list[vz.Trial],
    y_test: list[float],
) -> float:
  """Evaluate the designer's accuracy on the test set.

  Args:
    designer: The GP bandit designer to predict from.
    test_trials: The trials of the test set
    y_test: The results of the test set

  Returns:
    The MSE of `designer` on `test_trials` and `y_test`
  """
  preds = designer.predict(test_trials)
  return np.sum(np.square(preds.mean - y_test))


class GoogleGpBanditTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(iters=3, batch_size=2, num_seed_trials=1, ensemble_size=2),
      dict(
          iters=3,
          batch_size=1,
          num_seed_trials=1,
          ensemble_size=2,
          acquisition_optimizer_factory=lbfgsb_optimizer_factory,
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          iters=5,
          batch_size=1,
          num_seed_trials=3,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          acquisition_optimizer_factory=lbfgsb_optimizer_factory,
      ),
      dict(
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          ensemble_size=3,
      ),
  )
  def test_on_flat_continuous_space(
      self,
      *,
      iters: int = 5,
      batch_size: int = 1,
      num_seed_trials: int = 1,
      ensemble_size: int = 1,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(),
      use_trust_region: bool = False,
      acquisition_optimizer_factory: Union[
          vb.VectorizedOptimizerFactory, lo.LBFGSBOptimizerFactory
      ] = vectorized_optimizer_factory,
  ):
    # We use string names so that test case names are readable. Convert them
    # to objects.
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=acquisition_optimizer_factory,
        ard_optimizer=optimizers.JaxoptLbfgsB(
            optimizers.LbfgsBOptions(maxiter=5, num_line_search_steps=5)
        ),
        num_seed_trials=num_seed_trials,
        ensemble_size=ensemble_size,
        padding_schedule=padding_schedule,
        use_trust_region=use_trust_region,
        rng=jax.random.PRNGKey(0),
        linear_coef=0.1,
    )
    with profiler.collect_events() as events:
      self.assertLen(
          test_runners.RandomMetricsRunner(
              problem,
              iters=iters,
              batch_size=batch_size,
              verbose=1,
              validate_parameters=True,
              seed=1,
          ).run_designer(designer),
          iters * batch_size,
      )

    self.assertIn('VizierGPBandit.suggest', profiler.get_latencies_dict(events))

    quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        problem.search_space,
    )
    predict_trials = quasi_random_sampler.suggest(count=3)
    # test the sample method.
    samples = designer.sample(predict_trials, num_samples=5)
    self.assertSequenceEqual(samples.shape, (5, 3))
    self.assertFalse(np.isnan(samples).any())
    empty_samples = designer.sample([], num_samples=5)
    self.assertSequenceEqual(empty_samples.shape, (5, 0))
    # test the predict method.
    prediction = designer.predict(predict_trials)
    self.assertLen(prediction.mean, 3)
    self.assertLen(prediction.stddev, 3)
    self.assertFalse(np.isnan(prediction.mean).any())
    self.assertFalse(np.isnan(prediction.stddev).any())

  @parameterized.parameters(
      dict(iters=3, batch_size=5, num_seed_trials=5),
      dict(iters=5, batch_size=1, num_seed_trials=2),
      dict(iters=5, batch_size=1, num_seed_trials=1),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
  )
  def test_on_flat_mixed_space(
      self,
      iters: int,
      batch_size: int,
      num_seed_trials: int,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(),
      use_trust_region: bool = True,
  ):
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=num_seed_trials,
        padding_schedule=padding_schedule,
        use_trust_region=use_trust_region,
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            batch_size=batch_size,
            verbose=1,
            validate_parameters=True,
        ).run_designer(designer),
        iters * batch_size,
    )
    quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        problem.search_space
    )
    predict_trials = quasi_random_sampler.suggest(count=3)
    # Test the sample method.
    samples = designer.sample(predict_trials, num_samples=5)
    self.assertSequenceEqual(samples.shape, (5, 3))
    samples = designer.sample(predict_trials, num_samples=5)
    self.assertSequenceEqual(samples.shape, (5, 3))
    self.assertFalse(np.isnan(samples).any())
    empty_samples = designer.sample([], num_samples=5)
    self.assertSequenceEqual(empty_samples.shape, (5, 0))
    # Test the predict method.
    prediction = designer.predict(predict_trials)
    self.assertLen(prediction.mean, 3)
    self.assertLen(prediction.stddev, 3)
    self.assertFalse(np.isnan(prediction.mean).any())
    self.assertFalse(np.isnan(prediction.stddev).any())

  def test_invariance_to_trials_padding_on_flat_mixed_space(
      self,
  ):
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    iters = 5
    num_seed_trials = 1
    acquisition_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(),
        max_evaluations=100,
    )
    # TODO: The test fails with proper ARD. Fix that and then turn
    # on ARD again.
    noop_ard_optimizer = optimizers.default_optimizer(maxiter=0)
    desinger_rng = jax.random.PRNGKey(0)
    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=acquisition_optimizer_factory,
        ard_optimizer=noop_ard_optimizer,
        num_seed_trials=num_seed_trials,
        rng=desinger_rng,
    )
    padding_designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=acquisition_optimizer_factory,
        ard_optimizer=noop_ard_optimizer,
        num_seed_trials=num_seed_trials,
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.MULTIPLES_OF_10,
        ),
        rng=desinger_rng,
    )
    metrics_runner_seed = 1
    designer_suggestions = test_runners.RandomMetricsRunner(
        problem,
        iters=iters,
        verbose=1,
        seed=metrics_runner_seed,
        validate_parameters=True,
    ).run_designer(designer)

    padding_designer_suggestions = test_runners.RandomMetricsRunner(
        problem,
        iters=iters,
        verbose=1,
        seed=metrics_runner_seed,
        validate_parameters=True,
    ).run_designer(padding_designer)

    self.assertLen(designer_suggestions, iters)
    self.assertLen(padding_designer_suggestions, iters)
    for idx, (suggestion, padding_suggestion) in enumerate(
        zip(designer_suggestions, padding_designer_suggestions)
    ):
      params1 = suggestion.parameters.as_dict()
      params2 = padding_suggestion.parameters.as_dict()
      self.assertSameElements(params1.keys(), params2.keys())
      for key in params1.keys():
        self.assertAlmostEqual(
            params1[key],
            params2[key],
            places=5,
            msg=f'Mismatch in parameter: {key}, suggestions {idx}',
        )

  def test_prediction_accuracy(self):
    f = lambda x: -((x - 0.5) ** 2)
    gp_designer, obs_trials, _ = _setup_lambda_search(f)
    gp_designer.update(vza.CompletedTrials(obs_trials), vza.ActiveTrials())
    pred_trial = vz.Trial({'x0': 0.0})
    pred = gp_designer.predict([pred_trial])
    self.assertLess(np.abs(pred.mean[0] - f(0.0)), 2e-2)

  # TODO: Add assertions to this test. Ideally
  # create two designers with the same trial count (without padding)
  # padding is just an internal detail that should be tested separately.
  def test_jit_once(self, *args):
    del args
    jax.clear_caches()

    space = test_studies.flat_continuous_space_with_scaling()
    problem = vz.ProblemStatement(space)
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    def create_designer(problem):
      return gp_bandit.VizierGPBandit(
          problem=problem,
          acquisition_optimizer_factory=vectorized_optimizer_factory,
          num_seed_trials=3,
          ensemble_size=2,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
      )

    def create_runner(problem):
      return test_runners.RandomMetricsRunner(
          problem,
          iters=5,
          batch_size=1,
          verbose=1,
          validate_parameters=True,
      )

    designer = create_designer(problem)
    # Padding schedule should avoid retracing every iteration.
    create_runner(problem).run_designer(designer)

    # Padding schedule should avoid retracing with one more feature.
    space.root.add_float_param('x0', -5.0, 5.0)
    designer1 = create_designer(problem)
    create_runner(problem).run_designer(designer1)

    # Retracing should not occur when a new VizierGPBandit instance is created.
    designer2 = create_designer(problem)
    create_runner(problem).run_designer(designer2)

  def test_parallel_acquisition(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    def _qei_factory(data: types.ModelData) -> acquisitions.AcquisitionFunction:
      best_labels = acquisitions.get_best_labels(data.labels)
      return acquisitions.QEI(best_labels=best_labels, num_samples=100)

    scoring_fn_factory = acquisitions.bayesian_scoring_function_factory(
        _qei_factory
    )

    n_parallel = 4
    iters = 3
    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        ard_optimizer=optimizers.JaxoptLbfgsB(
            optimizers.LbfgsBOptions(maxiter=5, num_line_search_steps=5)
        ),
        scoring_function_factory=scoring_fn_factory,
        scoring_function_is_parallel=True,
        use_trust_region=False,
        num_seed_trials=n_parallel,
        ensemble_size=3,
        rng=jax.random.PRNGKey(0),
        linear_coef=0.1,
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            batch_size=n_parallel,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(designer),
        iters * n_parallel,
    )

  def test_multi_metrics(self):
    search_space = vz.SearchSpace()
    search_space.root.add_float_param('x0', -5.0, 5.0)
    problem = vz.ProblemStatement(
        search_space=search_space,
        metric_information=vz.MetricsConfig(
            metrics=[
                vz.MetricInformation(
                    'obj1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
                vz.MetricInformation(
                    'obj2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
            ]
        ),
    )

    iters = 2
    designer = gp_bandit.VizierGPBandit.from_problem(problem)
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(designer),
        iters,
    )


# TODO: Fix transfer learning and enable tests.
@unittest.skip('The current transfer learning seems broken and test failing.')
class GPBanditPriorsTest(parameterized.TestCase):

  def test_prior_warping(self):
    f = lambda x: -((x - 0.5) ** 2)
    transform_f = lambda x: -3 * ((x - 0.5) ** 2) + 10

    # X is in range of what is defined in `_setup_lambda_search`, [-5.0, 5.0)
    x_test = np.random.default_rng(1).uniform(-5.0, 5.0, 100)
    y_test = [transform_f(x) for x in x_test]
    test_trials = [vz.Trial({'x0': x}) for x in x_test]

    # Create the designer with a prior and the trials to train the prior.
    gp_designer_with_prior, obs_trials_for_prior, _ = _setup_lambda_search(
        f=f, num_trials=100
    )

    # Set priors to above trials.
    gp_designer_with_prior.set_priors(
        [vza.CompletedTrials(obs_trials_for_prior)]
    )

    # Create a no prior designer on the transformed function `transform_f`.
    # Also use the generated trials to update both the designer with prior and
    # the designer without. This tests that the prior designer is resilient
    # to linear transforms between the prior and the top level study.
    gp_designer_no_prior, obs_trials, _ = _setup_lambda_search(
        f=transform_f, num_trials=20
    )

    # Update both designers with the actual study.
    gp_designer_no_prior.update(
        vza.CompletedTrials(obs_trials), vza.ActiveTrials()
    )
    gp_designer_with_prior.update(
        vza.CompletedTrials(obs_trials), vza.ActiveTrials()
    )

    # Evaluate the no prior designer's accuracy on the test set.
    mse_no_prior = _compute_mse(gp_designer_no_prior, test_trials, y_test)

    # Evaluate the designer with prior's accuracy on the test set.
    mse_with_prior = _compute_mse(gp_designer_with_prior, test_trials, y_test)

    # The designer with a prior should predict better.
    self.assertLess(mse_with_prior, mse_no_prior)

  @parameterized.parameters(
      dict(iters=3, batch_size=5),
      dict(iters=5, batch_size=1),
  )
  def test_run_with_priors(self, *, iters, batch_size):
    f = lambda x: -((x - 0.5) ** 2)

    # Create the designer with a prior and the trials to train the prior.
    gp_designer_with_prior, obs_trials_for_prior, problem = (
        _setup_lambda_search(f=f, num_trials=100)
    )

    # Set priors to the above trials.
    gp_designer_with_prior.set_priors(
        [vza.CompletedTrials(obs_trials_for_prior)]
    )

    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            batch_size=batch_size,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(gp_designer_with_prior),
        iters * batch_size,
    )


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  jax.config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""OSS Vizier client."""

# TODO: Raise vizier-specific exceptions.

from typing import Any, Callable, Iterable, Iterator, List, Mapping, Optional, Type
import attr

from vizier._src.service import constants
from vizier._src.service import resources
from vizier._src.service import vizier_client
from vizier.client import client_abc
from vizier.service import pyvizier as vz

# Redeclared so users do not have to also import client_abc and clients.py.
ResourceNotFoundError = client_abc.ResourceNotFoundError

# Redeclared since clients.py is the user-facing client API.
environment_variables = vizier_client.environment_variables


@attr.define
class Trial(client_abc.TrialInterface):
  """Trial class.

  This class owns a Vizier client of the Study that contains the Trial that
  it is associated with.
  """

  _client: vizier_client.VizierClient = attr.field()
  _id: int = attr.field(validator=attr.validators.instance_of(int))

  @property
  def id(self) -> int:
    return self._id

  @property
  def parameters(self) -> Mapping[str, Any]:
    trial = self.materialize(include_all_measurements=False)
    study_config = self._client.get_study_config()
    return study_config.trial_parameters(vz.TrialConverter.to_proto(trial))

  def delete(self) -> None:
    self._client.delete_trial(self._id)

  def update_metadata(self, delta: vz.Metadata) -> None:
    actual_delta = vz.MetadataDelta(on_trials={self._id: delta})
    self._client.update_metadata(actual_delta)

  def complete(
      self,
      measurement: Optional[vz.Measurement] = None,
      *,
      infeasible_reason: Optional[str] = None,
  ) -> Optional[vz.Measurement]:
    self._trial = self._client.complete_trial(
        self._id, measurement, infeasible_reason
    )
    return self._trial.final_measurement

  def check_early_stopping(self) -> bool:
    return self._client.should_trial_stop(self._id)

  def stop(self) -> None:
    return self._client.stop_trial(self._id)

  def add_measurement(self, measurement: vz.Measurement) -> None:
    self._client.report_intermediate_objective_value(
        int(measurement.steps),
        measurement.elapsed_secs,
        [{k: v.value for k, v in measurement.metrics.items()}],
        trial_id=self._id,
    )

  def materialize(
      self,
      *,
      include_all_measurements: bool = True,
  ) -> vz.Trial:
    trial = self._client.get_trial(self._id)
    if not include_all_measurements:
      trial.measurements.clear()
    return trial

  @property
  def study(self) -> 'Study':
    return Study(self._client)


@attr.define
class TrialIterable(client_abc.TrialIterable):
  """Holds a collection of materialized Trials.

  See the parent class for full pydocs.
  """

  _iterable_factory: Callable[[], Iterable[vz.Trial]] = attr.field()
  _client: vizier_client.VizierClient = attr.field()

  def __iter__(self) -> Iterator[Trial]:
    for trial in self._iterable_factory():
      yield Trial(self._client, trial.id)

  def get(self) -> Iterator[vz.Trial]:
    for trial in self._iterable_factory():
      yield trial


@attr.define
class Study(client_abc.StudyInterface):
  """Responsible for study-level operations."""

  _client: vizier_client.VizierClient = attr.field()

  @property
  def resource_name(self) -> str:
    return self._client.study_resource_name

  def _trial_client(self, trial: vz.Trial) -> Trial:
    """Returns the client for the vz.Trial object."""
    return Trial(self._client, trial.id)

  def suggest(
      self, *, count: Optional[int] = None, client_id: str = 'default_client_id'
  ) -> List[Trial]:
    return [
        self._trial_client(t)
        for t in self._client.get_suggestions(
            count, client_id_override=client_id
        )
    ]

  def delete(self) -> None:
    self._client.delete_study()

  def update_metadata(self, delta: vz.Metadata) -> None:
    actual_delta = vz.MetadataDelta(on_study=delta)
    self._client.update_metadata(actual_delta)

  def _add_trial(self, trial: vz.Trial) -> Trial:
    return self._trial_client(self._client.add_trial(trial))

  def request(self, suggestion: vz.TrialSuggestion) -> None:
    trial = suggestion.to_trial()
    trial.is_requested = True
    self._client.add_trial(trial)

  def trials(
      self, trial_filter: Optional[vz.TrialFilter] = None
  ) -> TrialIterable:
    all_trials = self._client.list_trials()
    trial_filter = trial_filter or vz.TrialFilter()

    def iterable_factory():
      for t in filter(trial_filter, all_trials):
        yield t

    return TrialIterable(iterable_factory, self._client)

  def get_trial(self, trial_id: int) -> Trial:
    try:
      # Check if the trial actually exists.
      trial = self._client.get_trial(trial_id)
      return self._trial_client(trial)
    except KeyError as err:
      raise ResourceNotFoundError(
          f'Study f{self.resource_name} does not have Trial {trial_id}.'
      ) from err

  def optimal_trials(self) -> TrialIterable:
    trials = self._client.list_optimal_trials()
    return TrialIterable(lambda: trials, self._client)

  def materialize_problem_statement(self) -> vz.ProblemStatement:
    """Returns a cross-platform compatible minimal StudyConfig."""
    return self._client.get_study_config().to_problem()

  def materialize_study_config(self) -> vz.StudyConfig:
    """Returns a fully specific StudyConfig."""
    return self._client.get_study_config(self.resource_name)

  def set_state(self, state: vz.StudyState) -> None:
    self._client.set_study_state(state)

  def materialize_state(self) -> vz.StudyState:
    return self._client.get_study_state()

  @classmethod
  def from_resource_name(cls: Type['Study'], name: str) -> 'Study':
    client = vizier_client.VizierClient(
        name,
        constants.UNUSED_CLIENT_ID,
    )
    try:
      _ = client.get_study_config()  # Make sure study exists.
    except Exception as err:
      raise KeyError(f'Study {name} does not exist.') from err
    return Study(client)

  @classmethod
  def from_owner_and_id(
      cls: Type['Study'], owner: str, study_id: str
  ) -> 'Study':
    """Create study from `owner` and `study_id`.

    Args:
      owner: Owner of the study.
      study_id: Unique identifier within the same owner.

    Returns:
      Study.

    Raises:
      ResourceNotFoundError.
    """
    study_resource_name = resources.StudyResource(
        owner_id=owner, study_id=study_id
    ).name
    return cls.from_resource_name(study_resource_name)

  @classmethod
  def from_study_config(
      cls, config: vz.StudyConfig, *, owner: str, study_id: str
  ) -> 'Study':
    """Create study from StudyConfig.

    Args:
      config: OSS Study configuration. It is platform-specific. It is ignored if
        `owner` already has Study with `study_id`. # TODO: Instead
        of ignoring it, compare with the existing # study and return error if
        there's a change.
      owner: Owner of the study.
      study_id: Unique identifier within the same owner.

    Returns:
      Study.
    """
    # Use a dummy client_id, since for multi-worker workflows, each worker is
    # expected to provide a client_id in the suggest() call.
    return Study(
        vizier_client.create_or_load_study(
            owner_id=owner,
            client_id=constants.UNUSED_CLIENT_ID,
            study_id=study_id,
            study_config=config,
        )
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import random

from scipy import stats
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.policies import designer_policy
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class QuasiRandomTest(absltest.TestCase):

  def test_on_flat_space(self):
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = quasi_random.QuasiRandomDesigner(problem.search_space)
    self.assertLen(
        test_runners.run_with_random_metrics(
            designer, problem, iters=50, batch_size=5, validate_parameters=True
        ),
        250,
    )

  def test_dump_and_load(self):
    # Check metadata checkpointing.
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = quasi_random.QuasiRandomDesigner(problem.search_space)
    designer.suggest(random.randint(1, 50))
    metadata = designer.dump()
    designer2 = quasi_random.QuasiRandomDesigner(problem.search_space)
    designer2.load(metadata)
    self.assertEqual(designer.suggest(10), designer2.suggest(10))

  def test_distribution(self):
    # Make sure output distribution makes sense.
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer = quasi_random.QuasiRandomDesigner(problem.search_space)

    suggestions = designer.suggest(2000)
    float_points = [
        suggestion.parameters['float'].value for suggestion in suggestions
    ]

    # Test uniformity of end-to-end parameter values.
    # p_value greater than 0.9 roughly means we're very certain it's uniform.
    # Unfortunately KS-test doesn't work for discrete/categorical distributions.
    _, float_p_value = stats.kstest(
        float_points, stats.uniform(loc=0.0, scale=1.0).cdf
    )
    self.assertGreater(float_p_value, 0.9)

  def test_equal_seeds(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer_1 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_1 = designer_1.suggest(10)
    designer_2 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_2 = designer_2.suggest(10)
    self.assertEqual(suggestions_1, suggestions_2)

  def test_distinct_seeds(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer_1 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=0)
    suggestions_1 = designer_1.suggest(10)
    designer_2 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_2 = designer_2.suggest(10)
    self.assertNotEqual(suggestions_1, suggestions_2)

  def test_policy_wrapping(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    policy_supporter = pythia.InRamPolicySupporter(problem)
    policy = designer_policy.PartiallySerializableDesignerPolicy(
        problem,
        policy_supporter,
        quasi_random.QuasiRandomDesigner.from_problem,
    )

    # Make sure outputs are distinct.
    all_suggestions = []
    for _ in range(1000):
      request = pythia.SuggestRequest(
          study_descriptor=policy_supporter.study_descriptor(), count=1
      )
      decisions = policy.suggest(request)
      all_suggestions.extend(decisions.suggestions)

    distinct_suggestions = set(
        [
            tuple(suggestion.parameters.as_dict().values())
            for suggestion in all_suggestions
        ]
    )
    self.assertLen(distinct_suggestions, 1000)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gradient-free optimizers."""

from absl import logging
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz


def assert_passes_on_random_single_metric_function(
    self, search_space: vz.SearchSpace, optimizer: vza.GradientFreeOptimizer, *,
    np_random_seed: int):
  """Smoke test on random score."""
  rng = np.random.default_rng(np_random_seed)

  logging.info('search space: %s', search_space)

  problem = vz.ProblemStatement(
      search_space=search_space,
      metric_information=[
          vz.MetricInformation(
              'acquisition', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
      ])

  def mock_score(trials):
    return {'acquisition': rng.uniform(size=[len(trials), 1])}

  suggestions = optimizer.optimize(mock_score, problem, count=5)
  self.assertNotEmpty(suggestions)

  logging.info('suggestions: %s', suggestions)
  for suggestion in suggestions:
    problem.search_space.assert_contains(suggestion.parameters)


def assert_passes_on_random_multi_metric_function(
    self,
    search_space: vz.SearchSpace,
    optimizer: vza.GradientFreeOptimizer,
    *,
    np_random_seed: int
):
  """Bi-objective test on random score."""
  rng = np.random.default_rng(np_random_seed)

  logging.info('search space: %s', search_space)

  problem = vz.ProblemStatement(
      search_space=search_space,
      metric_information=[
          vz.MetricInformation(
              'acquisition_1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
          ),
          vz.MetricInformation(
              'acquisition_2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
          ),
      ],
  )

  def mock_score(trials):
    return {
        'acquisition_1': rng.uniform(size=[len(trials), 1]),
        'acquisition_2': rng.uniform(size=[len(trials), 1]),
    }

  suggestions = optimizer.optimize(mock_score, problem, count=5)
  self.assertNotEmpty(suggestions)

  logging.info('suggestions: %s', suggestions)
  for suggestion in suggestions:
    problem.search_space.assert_contains(suggestion.parameters)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import multiobjective_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest


class MultiobjectiveExperimenterTest(absltest.TestCase):

  def test_mulitobjective_numpy(self):
    dim = 2
    func1 = bbob.Sphere
    func2 = bbob.Rastrigin
    exptr1 = numpy_experimenter.NumpyExperimenter(
        func1, bbob.DefaultBBOBProblemStatement(dim)
    )
    exptr2 = numpy_experimenter.NumpyExperimenter(
        func2, bbob.DefaultBBOBProblemStatement(dim)
    )
    exptr = multiobjective_experimenter.MultiObjectiveExperimenter(
        {'m1': exptr1, 'm2': exptr2}
    )
    parameters = exptr1.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(
        parameters={
            param.name: float(index) for index, param in enumerate(parameters)
        }
    )

    exptr.evaluate([t])
    self.assertAlmostEqual(
        func1(np.array([0.0, 1.0])),
        t.final_measurement.metrics['m1'].value,
    )
    self.assertAlmostEqual(
        func2(np.array([0.0, 1.0])),
        t.final_measurement.metrics['m2'].value,
    )
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)

  def test_dimension_mismatch(self):
    exptr1 = numpy_experimenter.NumpyExperimenter(
        bbob.Sphere, bbob.DefaultBBOBProblemStatement(dimension=3)
    )
    exptr2 = numpy_experimenter.NumpyExperimenter(
        bbob.Rastrigin, bbob.DefaultBBOBProblemStatement(dimension=6)
    )
    with self.assertRaisesRegex(ValueError, 'space must match'):
      multiobjective_experimenter.MultiObjectiveExperimenter(
          {'m1': exptr1, 'm2': exptr2}
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Vizier problem statement and trials scaler.

The ProblemAndTrialsScaler class has two main functionalities:
1. Scale (aka "embed") the ProblemStatment to a new ProblemStatment with scaled
search space.
2. Map and unmap Vizier TrialSuggestion from and to the scaled search space.

The parameters mapping is as follow:
- FLOAT parameters are mapped to FLOAT with [0,1] bounds.
- INTEGER parameters are mapped to FLOAT with [0,1] bounds.
- DISCRETE parameters are mapped to DISCRETE with converted feasible values.
- CATEGORICAL parameters don't change.

The scaling is performed using the DefaultTrialConverter.
"""

import copy
from typing import Sequence, TypeVar
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier.pyvizier.converters import core


_T = TypeVar('_T', vz.Trial, vz.TrialSuggestion)


class ProblemAndTrialsScaler:
  """Vizier problem statement and trials scaler."""

  def __init__(self, problem: vz.ProblemStatement):
    """Initializes the class by creating a converter and scaled problem statement.

    Arguments:
      problem: The ProblemStatemet to be scaled.
    """

    def create_param_converter(pc):
      return core.DefaultModelInputConverter(
          pc, max_discrete_indices=0, scale=True
      )

    self._converter = core.DefaultTrialConverter(
        parameter_converters=[
            create_param_converter(pc) for pc in problem.search_space.parameters
        ]
    )
    # Create the new embedded search space.
    emb_search_space = vz.SearchSpace()
    for param in problem.search_space.parameters:
      if param.type in [vz.ParameterType.DOUBLE, vz.ParameterType.INTEGER]:
        # DOUBLE/INTEGER params are scaled to [0.0, 1.0] and converted to FLOAT.
        emb_search_space.root.add_float_param(
            param.name,
            self._converter.output_specs[param.name].bounds[0],
            self._converter.output_specs[param.name].bounds[1],
        )
      elif param.type == vz.ParameterType.DISCRETE:
        # DISCRETE params feasible values are scaled.
        emb_search_space.root.add_discrete_param(
            param.name,
            feasible_values=self._scale_discrete_feasible_values(
                problem, param
            ),
        )
      elif param.type == vz.ParameterType.CATEGORICAL:
        # CATEGORICAL parameters are left unchanged.
        emb_search_space.root.add_categorical_param(
            param.name, feasible_values=param.feasible_values
        )
      else:
        raise ValueError('Unsupported parameter type (%s)' % param.type)
    # Clone the entire problem statement and only update the search space.
    self._embedded_problem_statement = copy.deepcopy(problem)
    self._embedded_problem_statement.search_space = emb_search_space

  @property
  def problem_statement(self) -> vz.ProblemStatement:
    """Returns the scaled problem statement."""
    return self._embedded_problem_statement

  def _scale_discrete_feasible_values(
      self, problem: vz.ProblemStatement, param: vz.ParameterConfig
  ):
    """Scales the DISCRETE parameter feasible values."""
    if param.type != vz.ParameterType.DISCRETE:
      raise ValueError('Expects DISCRETE parameter but got %s.' % param.type)
    # Obtain the converter for the DISCRETE parameter.
    param_converter = self._converter.parameter_converters_dict[param.name]
    # Convert each original feasible values to form the new feasible values.
    new_feasible_values = []
    for feasible_value in problem.search_space.get(param.name).feasible_values:
      tmp_trial = vz.Trial({param.name: feasible_value})
      new_feasible_value = param_converter.convert([tmp_trial]).item(0, 0)
      new_feasible_values.append(new_feasible_value)
    return new_feasible_values

  def map(self, trials: Sequence[_T]) -> list[_T]:
    """Map trials from the original search space to the embedded space."""
    embedded_trials = []
    for trial in trials:
      parameters = vz.ParameterDict()
      for name, feature in self._converter.to_features([trial]).items():
        if (
            self.problem_statement.search_space.get(name).type
            == vz.ParameterType.CATEGORICAL
        ):
          # CATEGORICAL parameters values are unchanged.
          parameters[name] = trial.parameters[name].value
        else:
          # Assign the converted value.
          parameters[name] = feature.item(0, 0)
      # Create a copy of the trial with updated parameters.
      embedded_trials.append(attr.evolve(trial, parameters=parameters))

    return embedded_trials

  def unmap(self, trials: Sequence[_T]) -> list[_T]:
    """Unmap trials from the embedded search space to the original space."""
    unmapped_trials = []
    for trial in trials:
      parameters = vz.ParameterDict()
      for name, parameter_value in trial.parameters.items():
        value = parameter_value.value
        if (
            self.problem_statement.search_space.get(name).type
            == vz.ParameterType.CATEGORICAL
        ):
          # CATEGORICAL parameter values are left unchanged.
          parameters[name] = value
        else:
          # Get the parameter converter.
          param_converter = self._converter.parameter_converters_dict[name]
          # Convert back the feature to parameters in the original space.
          parameters[name] = param_converter.to_parameter_values(
              np.array(value)
          )[0]
      # Create a copy of the trial with updated parameters.
      unmapped_trials.append(attr.evolve(trial, parameters=parameters))
    return unmapped_trials



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Policy Supporter used for the OSS Vizier service.

The Policy can use these methods to communicate with Vizier.
"""
import datetime
from typing import Iterable, List, Optional

from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.service import types
from vizier._src.service import vizier_service_pb2
from vizier.service import pyvizier


# TODO: Consider replacing protos with Pyvizier clients.py.
class ServicePolicySupporter(pythia.PolicySupporter):
  """Service version of the PolicySupporter."""

  def __init__(self, study_guid: str, vizier_service: types.VizierService):
    """Initalization.

    Args:
      study_guid: A default study_name; the name of this study.
      vizier_service: Vizier Service, in the form of a GRPC stub or actual
        class.
    """
    self._study_guid = study_guid
    self._vizier_service = vizier_service

  @property
  def study_guid(self) -> str:
    return self._study_guid

  def GetStudyConfig(self, study_guid: str) -> vz.ProblemStatement:
    request = vizier_service_pb2.GetStudyRequest(name=study_guid)
    study = self._vizier_service.GetStudy(request)
    return pyvizier.StudyConfig.from_proto(study.study_spec).to_problem()

  # TODO: Support filters in ListTrialsRequest.
  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,  # Equivalent to Study.name in OSS.
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True,
  ) -> List[vz.Trial]:
    """Fetch all trials and then apply the filter."""

    if study_guid is None:
      study_guid = self._study_guid
    request = vizier_service_pb2.ListTrialsRequest(parent=study_guid)
    trials = self._vizier_service.ListTrials(request).trials
    all_pytrials = pyvizier.TrialConverter.from_protos(trials)

    trial_filter = vz.TrialFilter(
        ids=trial_ids,
        min_id=min_trial_id,
        max_id=max_trial_id,
        status=[status_matches] if status_matches else None,
    )
    filtered_pytrials = [t for t in all_pytrials if trial_filter(t)]

    # Doesn't affect datastore when measurements are deleted.
    if not include_intermediate_measurements:
      for filtered_pytrial in filtered_pytrials:
        filtered_pytrial.measurements = []

    return filtered_pytrials

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    """Throws a CancelComputeError on timeout or if Vizier cancels."""
    pass  # Do nothing since it's one single process.

  def TimeRemaining(self) -> datetime.timedelta:
    """The time remaining to compute a result."""
    return datetime.timedelta.max  # RPCs don't have timeouts in OSS.



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Random Pythia Policy which produces uniform sampling of Trial parameter values.

Since this is a RandomPolicy (i.e. stateless), we don't use the PolicySupporter
API when suggesting trials, but we do for the early stopping in order to
showcase how the policy supporter should be used.
"""
import random
from vizier import pythia
from vizier import pyvizier
from vizier._src.algorithms.designers import random as random_designer_lib


class RandomPolicy(pythia.Policy):
  """A policy that picks random hyper-parameter values."""

  def __init__(self, policy_supporter: pythia.PolicySupporter):
    self._policy_supporter = policy_supporter

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    """Gets number of Trials to propose, and produces random Trials."""
    suggestions = random_designer_lib.RandomDesigner(
        request.study_config.search_space
    ).suggest(request.count)
    return pythia.SuggestDecision(suggestions)

  def early_stop(
      self, request: pythia.EarlyStopRequest
  ) -> pythia.EarlyStopDecisions:
    """Selects a random ACTIVE/PENDING trial to stop from datastore."""
    decisions = []

    all_active_trials = self._policy_supporter.GetTrials(
        study_guid=request.study_guid,
        status_matches=pyvizier.TrialStatus.ACTIVE,
    )
    trial_to_stop_id = None
    if all_active_trials:
      trial_to_stop_id = random.choice(all_active_trials).id
      decisions.append(
          pythia.EarlyStopDecision(
              id=trial_to_stop_id, reason='Random early stopping.'
          )
      )

    for trial_id in list(request.trial_ids):
      if trial_id != trial_to_stop_id:
        decisions.append(
            pythia.EarlyStopDecision(
                id=trial_id, reason='Trial should not stop.', should_stop=False
            )
        )

    return pythia.EarlyStopDecisions(decisions, pyvizier.MetadataDelta())
